@@ -26,6 +26,7 @@
 #include "ovn-controller.h"
 #include "ovn/actions.h"
 #include "ovn/expr.h"
+#include "lib/lb.h"
 #include "lib/ovn-l7.h"
 #include "lib/ovn-sb-idl.h"
 #include "lib/extend-table.h"
@@ -666,6 +667,7 @@ update_conj_id_ofs(uint32_t *conj_id_ofs, uint32_t n_conjs)
 
 static void
 add_matches_to_flow_table(const struct sbrec_logical_flow *lflow,
+                          const struct sbrec_datapath_binding *dp,
                           struct hmap *matches, size_t conj_id_ofs,
                           uint8_t ptable, uint8_t output_ptable,
                           struct ofpbuf *ovnacts,
@@ -676,7 +678,7 @@ add_matches_to_flow_table(const struct sbrec_logical_flow *lflow,
         .sbrec_multicast_group_by_name_datapath
             = l_ctx_in->sbrec_multicast_group_by_name_datapath,
         .sbrec_port_binding_by_name = l_ctx_in->sbrec_port_binding_by_name,
-        .dp = lflow->logical_datapath
+        .dp = dp,
     };
 
     /* Encode OVN logical actions into OpenFlow. */
@@ -686,7 +688,7 @@ add_matches_to_flow_table(const struct sbrec_logical_flow *lflow,
         .lookup_port = lookup_port_cb,
         .tunnel_ofport = tunnel_ofport_cb,
         .aux = &aux,
-        .is_switch = datapath_is_switch(lflow->logical_datapath),
+        .is_switch = datapath_is_switch(dp),
         .group_table = l_ctx_out->group_table,
         .meter_table = l_ctx_out->meter_table,
         .lflow_uuid = lflow->header_.uuid,
@@ -697,22 +699,24 @@ add_matches_to_flow_table(const struct sbrec_logical_flow *lflow,
         .output_ptable = output_ptable,
         .mac_bind_ptable = OFTABLE_MAC_BINDING,
         .mac_lookup_ptable = OFTABLE_MAC_LOOKUP,
+        .lb_hairpin_ptable = OFTABLE_CHK_LB_HAIRPIN,
+        .lb_hairpin_reply_ptable = OFTABLE_CHK_LB_HAIRPIN_REPLY,
+        .ct_snat_vip_ptable = OFTABLE_CT_SNAT_FOR_VIP,
     };
     ovnacts_encode(ovnacts->data, ovnacts->size, &ep, &ofpacts);
 
     struct expr_match *m;
     HMAP_FOR_EACH (m, hmap_node, matches) {
-        match_set_metadata(&m->match,
-                           htonll(lflow->logical_datapath->tunnel_key));
+        match_set_metadata(&m->match, htonll(dp->tunnel_key));
         if (m->match.wc.masks.conj_id) {
             m->match.flow.conj_id += conj_id_ofs;
         }
-        if (datapath_is_switch(lflow->logical_datapath)) {
+        if (datapath_is_switch(dp)) {
             unsigned int reg_index
                 = (ingress ? MFF_LOG_INPORT : MFF_LOG_OUTPORT) - MFF_REG0;
             int64_t port_id = m->match.flow.regs[reg_index];
             if (port_id) {
-                int64_t dp_id = lflow->logical_datapath->tunnel_key;
+                int64_t dp_id = dp->tunnel_key;
                 char buf[16];
                 get_unique_lport_key(dp_id, port_id, buf, sizeof(buf));
                 lflow_resource_add(l_ctx_out->lfrr, REF_TYPE_PORTBINDING, buf,
@@ -755,41 +759,48 @@ add_matches_to_flow_table(const struct sbrec_logical_flow *lflow,
     ofpbuf_uninit(&ofpacts);
 }
 
-/* Converts the match and returns the simplified expre tree.
- * The caller should evaluate the conditions and normalize the
- * expr tree. */
+/* Converts the match and returns the simplified expr tree.
+ *
+ * The caller should evaluate the conditions and normalize the expr tree.
+ */
 static struct expr *
 convert_match_to_expr(const struct sbrec_logical_flow *lflow,
+                      const struct sbrec_datapath_binding *dp,
                       struct expr *prereqs,
-                      struct lflow_ctx_in *l_ctx_in,
-                      struct lflow_ctx_out *l_ctx_out,
-                      bool *pg_addr_set_ref, char **errorp)
+                      const struct shash *addr_sets,
+                      const struct shash *port_groups,
+                      struct lflow_resource_ref *lfrr,
+                      bool *pg_addr_set_ref)
 {
     struct sset addr_sets_ref = SSET_INITIALIZER(&addr_sets_ref);
     struct sset port_groups_ref = SSET_INITIALIZER(&port_groups_ref);
-    char *error;
+    char *error = NULL;
 
-    struct expr *e = expr_parse_string(lflow->match, &symtab,
-                                       l_ctx_in->addr_sets,
-                                       l_ctx_in->port_groups, &addr_sets_ref,
-                                       &port_groups_ref,
-                                       lflow->logical_datapath->tunnel_key,
+    struct expr *e = expr_parse_string(lflow->match, &symtab, addr_sets,
+                                       port_groups, &addr_sets_ref,
+                                       &port_groups_ref, dp->tunnel_key,
                                        &error);
     const char *addr_set_name;
     SSET_FOR_EACH (addr_set_name, &addr_sets_ref) {
-        lflow_resource_add(l_ctx_out->lfrr, REF_TYPE_ADDRSET, addr_set_name,
+        lflow_resource_add(lfrr, REF_TYPE_ADDRSET, addr_set_name,
                            &lflow->header_.uuid);
     }
     const char *port_group_name;
     SSET_FOR_EACH (port_group_name, &port_groups_ref) {
-        lflow_resource_add(l_ctx_out->lfrr, REF_TYPE_PORTGROUP,
-                           port_group_name, &lflow->header_.uuid);
+        lflow_resource_add(lfrr, REF_TYPE_PORTGROUP, port_group_name,
+                           &lflow->header_.uuid);
     }
 
+    if (pg_addr_set_ref) {
+        *pg_addr_set_ref = (!sset_is_empty(&port_groups_ref) ||
+                            !sset_is_empty(&addr_sets_ref));
+    }
+    sset_destroy(&addr_sets_ref);
+    sset_destroy(&port_groups_ref);
+
     if (!error) {
         if (prereqs) {
             e = expr_combine(EXPR_T_AND, e, prereqs);
-            prereqs = NULL;
         }
         e = expr_annotate(e, &symtab, &error);
     }
@@ -797,44 +808,26 @@ convert_match_to_expr(const struct sbrec_logical_flow *lflow,
         static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(1, 1);
         VLOG_WARN_RL(&rl, "error parsing match \"%s\": %s",
                     lflow->match, error);
-        sset_destroy(&addr_sets_ref);
-        sset_destroy(&port_groups_ref);
-        *errorp = error;
+        free(error);
         return NULL;
     }
 
-    e = expr_simplify(e);
-
-    if (pg_addr_set_ref) {
-        *pg_addr_set_ref = (!sset_is_empty(&port_groups_ref) ||
-                            !sset_is_empty(&addr_sets_ref));
-    }
-
-    sset_destroy(&addr_sets_ref);
-    sset_destroy(&port_groups_ref);
-
-    *errorp = NULL;
-    return e;
+    return expr_simplify(e);
 }
 
 static bool
-consider_logical_flow(const struct sbrec_logical_flow *lflow,
-                      struct hmap *dhcp_opts, struct hmap *dhcpv6_opts,
-                      struct hmap *nd_ra_opts,
-                      struct controller_event_options *controller_event_opts,
-                      struct lflow_ctx_in *l_ctx_in,
-                      struct lflow_ctx_out *l_ctx_out)
+consider_logical_flow__(const struct sbrec_logical_flow *lflow,
+                        const struct sbrec_datapath_binding *dp,
+                        struct hmap *dhcp_opts, struct hmap *dhcpv6_opts,
+                        struct hmap *nd_ra_opts,
+                        struct controller_event_options *controller_event_opts,
+                        struct lflow_ctx_in *l_ctx_in,
+                        struct lflow_ctx_out *l_ctx_out)
 {
     /* Determine translation of logical table IDs to physical table IDs. */
     bool ingress = !strcmp(lflow->pipeline, "ingress");
 
-    const struct sbrec_datapath_binding *ldp = lflow->logical_datapath;
-    if (!ldp) {
-        VLOG_DBG("lflow "UUID_FMT" has no datapath binding, skip",
-                 UUID_ARGS(&lflow->header_.uuid));
-        return true;
-    }
-    if (!get_local_datapath(l_ctx_in->local_datapaths, ldp->tunnel_key)) {
+    if (!get_local_datapath(l_ctx_in->local_datapaths, dp->tunnel_key)) {
         VLOG_DBG("lflow "UUID_FMT" is not for local datapath, skip",
                  UUID_ARGS(&lflow->header_.uuid));
         return true;
@@ -883,7 +876,7 @@ consider_logical_flow(const struct sbrec_logical_flow *lflow,
         .sbrec_multicast_group_by_name_datapath
             = l_ctx_in->sbrec_multicast_group_by_name_datapath,
         .sbrec_port_binding_by_name = l_ctx_in->sbrec_port_binding_by_name,
-        .dp = lflow->logical_datapath
+        .dp = dp,
     };
     struct condition_aux cond_aux = {
         .sbrec_port_binding_by_name = l_ctx_in->sbrec_port_binding_by_name,
@@ -896,13 +889,13 @@ consider_logical_flow(const struct sbrec_logical_flow *lflow,
     struct expr *expr = NULL;
     if (!l_ctx_out->lflow_cache_map) {
         /* Caching is disabled. */
-        expr = convert_match_to_expr(lflow, prereqs, l_ctx_in,
-                                     l_ctx_out, NULL, &error);
-        if (error) {
+        expr = convert_match_to_expr(lflow, dp, prereqs, l_ctx_in->addr_sets,
+                                     l_ctx_in->port_groups, l_ctx_out->lfrr,
+                                     NULL);
+        if (!expr) {
             expr_destroy(prereqs);
             ovnacts_free(ovnacts.data, ovnacts.size);
             ofpbuf_uninit(&ovnacts);
-            free(error);
             return true;
         }
 
@@ -922,7 +915,7 @@ consider_logical_flow(const struct sbrec_logical_flow *lflow,
             return true;
         }
 
-        add_matches_to_flow_table(lflow, &matches, *l_ctx_out->conj_id_ofs,
+        add_matches_to_flow_table(lflow, dp, &matches, *l_ctx_out->conj_id_ofs,
                                   ptable, output_ptable, &ovnacts, ingress,
                                   l_ctx_in, l_ctx_out);
 
@@ -939,7 +932,7 @@ consider_logical_flow(const struct sbrec_logical_flow *lflow,
     if (lc && lc->type == LCACHE_T_MATCHES) {
         /* 'matches' is cached. No need to do expr parsing.
          * Add matches to flow table and return. */
-        add_matches_to_flow_table(lflow, lc->expr_matches, lc->conj_id_ofs,
+        add_matches_to_flow_table(lflow, dp, lc->expr_matches, lc->conj_id_ofs,
                                   ptable, output_ptable, &ovnacts, ingress,
                                   l_ctx_in, l_ctx_out);
         ovnacts_free(ovnacts.data, ovnacts.size);
@@ -959,13 +952,13 @@ consider_logical_flow(const struct sbrec_logical_flow *lflow,
 
     bool pg_addr_set_ref = false;
     if (!expr) {
-        expr = convert_match_to_expr(lflow, prereqs, l_ctx_in, l_ctx_out,
-                                     &pg_addr_set_ref, &error);
-        if (error) {
+        expr = convert_match_to_expr(lflow, dp, prereqs, l_ctx_in->addr_sets,
+                                     l_ctx_in->port_groups, l_ctx_out->lfrr,
+                                     &pg_addr_set_ref);
+        if (!expr) {
             expr_destroy(prereqs);
             ovnacts_free(ovnacts.data, ovnacts.size);
             ofpbuf_uninit(&ovnacts);
-            free(error);
             return true;
         }
     } else {
@@ -1017,7 +1010,7 @@ consider_logical_flow(const struct sbrec_logical_flow *lflow,
     }
 
     /* Encode OVN logical actions into OpenFlow. */
-    add_matches_to_flow_table(lflow, matches, lc->conj_id_ofs,
+    add_matches_to_flow_table(lflow, dp, matches, lc->conj_id_ofs,
                               ptable, output_ptable, &ovnacts, ingress,
                               l_ctx_in, l_ctx_out);
     ovnacts_free(ovnacts.data, ovnacts.size);
@@ -1039,6 +1032,42 @@ consider_logical_flow(const struct sbrec_logical_flow *lflow,
     return true;
 }
 
+static bool
+consider_logical_flow(const struct sbrec_logical_flow *lflow,
+                      struct hmap *dhcp_opts, struct hmap *dhcpv6_opts,
+                      struct hmap *nd_ra_opts,
+                      struct controller_event_options *controller_event_opts,
+                      struct lflow_ctx_in *l_ctx_in,
+                      struct lflow_ctx_out *l_ctx_out)
+{
+    const struct sbrec_logical_dp_group *dp_group = lflow->logical_dp_group;
+    const struct sbrec_datapath_binding *dp = lflow->logical_datapath;
+    bool ret = true;
+
+    if (!dp_group && !dp) {
+        VLOG_DBG("lflow "UUID_FMT" has no datapath binding, skip",
+                 UUID_ARGS(&lflow->header_.uuid));
+        return true;
+    }
+    ovs_assert(!dp_group || !dp);
+
+    if (dp && !consider_logical_flow__(lflow, dp,
+                                       dhcp_opts, dhcpv6_opts, nd_ra_opts,
+                                       controller_event_opts,
+                                       l_ctx_in, l_ctx_out)) {
+        ret = false;
+    }
+    for (size_t i = 0; dp_group && i < dp_group->n_datapaths; i++) {
+        if (!consider_logical_flow__(lflow, dp_group->datapaths[i],
+                                     dhcp_opts,  dhcpv6_opts, nd_ra_opts,
+                                     controller_event_opts,
+                                     l_ctx_in, l_ctx_out)) {
+            ret = false;
+        }
+    }
+    return ret;
+}
+
 static void
 put_load(const uint8_t *data, size_t len,
          enum mf_field_id dst, int ofs, int n_bits,
@@ -1144,6 +1173,203 @@ add_neighbor_flows(struct ovsdb_idl_index *sbrec_port_binding_by_name,
     }
 }
 
+static void
+add_lb_vip_hairpin_flows(struct ovn_controller_lb *lb,
+                         struct ovn_lb_vip *lb_vip,
+                         struct ovn_lb_backend *lb_backend,
+                         uint8_t lb_proto,
+                         struct ovn_desired_flow_table *flow_table)
+{
+    uint64_t stub[1024 / 8];
+    struct ofpbuf ofpacts = OFPBUF_STUB_INITIALIZER(stub);
+
+    uint8_t value = 1;
+    put_load(&value, sizeof value, MFF_LOG_FLAGS,
+             MLF_LOOKUP_LB_HAIRPIN_BIT, 1, &ofpacts);
+
+    struct match hairpin_match = MATCH_CATCHALL_INITIALIZER;
+    struct match hairpin_reply_match = MATCH_CATCHALL_INITIALIZER;
+
+    if (IN6_IS_ADDR_V4MAPPED(&lb_vip->vip)) {
+        ovs_be32 ip4 = in6_addr_get_mapped_ipv4(&lb_backend->ip);
+
+        match_set_dl_type(&hairpin_match, htons(ETH_TYPE_IP));
+        match_set_nw_src(&hairpin_match, ip4);
+        match_set_nw_dst(&hairpin_match, ip4);
+
+        match_set_dl_type(&hairpin_reply_match,
+                          htons(ETH_TYPE_IP));
+        match_set_nw_src(&hairpin_reply_match, ip4);
+        match_set_nw_dst(&hairpin_reply_match,
+                         in6_addr_get_mapped_ipv4(&lb_vip->vip));
+    } else {
+        match_set_dl_type(&hairpin_match, htons(ETH_TYPE_IPV6));
+        match_set_ipv6_src(&hairpin_match, &lb_backend->ip);
+        match_set_ipv6_dst(&hairpin_match, &lb_backend->ip);
+
+        match_set_dl_type(&hairpin_reply_match,
+                          htons(ETH_TYPE_IPV6));
+        match_set_ipv6_src(&hairpin_reply_match, &lb_backend->ip);
+        match_set_ipv6_dst(&hairpin_reply_match, &lb_vip->vip);
+    }
+
+    if (lb_backend->port) {
+        match_set_nw_proto(&hairpin_match, lb_proto);
+        match_set_tp_dst(&hairpin_match, htons(lb_backend->port));
+
+        match_set_nw_proto(&hairpin_reply_match, lb_proto);
+        match_set_tp_src(&hairpin_reply_match, htons(lb_backend->port));
+    }
+
+    /* In the original direction, only match on traffic that was already
+     * load balanced, i.e., "ct.natted == 1".  Also, it's good enough
+     * to not include the datapath tunnel_key in the match when determining
+     * that a packet needs to be hairpinned because the rest of the match is
+     * restrictive enough:
+     * - traffic must have already been load balanced.
+     * - packets must have ip.src == ip.dst at this point.
+     * - the destination protocol and port must be of a valid backend that
+     *   has the same IP as ip.dst.
+     */
+    ovs_u128 lb_ct_label = {
+        .u64.lo = OVN_CT_NATTED,
+    };
+    match_set_ct_label_masked(&hairpin_match, lb_ct_label, lb_ct_label);
+
+    ofctrl_add_flow(flow_table, OFTABLE_CHK_LB_HAIRPIN, 100,
+                    lb->slb->header_.uuid.parts[0], &hairpin_match,
+                    &ofpacts, &lb->slb->header_.uuid);
+
+    for (size_t i = 0; i < lb->slb->n_datapaths; i++) {
+        match_set_metadata(&hairpin_reply_match,
+                           htonll(lb->slb->datapaths[i]->tunnel_key));
+
+        ofctrl_add_flow(flow_table, OFTABLE_CHK_LB_HAIRPIN_REPLY, 100,
+                        lb->slb->header_.uuid.parts[0],
+                        &hairpin_reply_match,
+                        &ofpacts, &lb->slb->header_.uuid);
+    }
+
+    ofpbuf_uninit(&ofpacts);
+}
+
+static void
+add_lb_ct_snat_vip_flows(struct ovn_controller_lb *lb,
+                         struct ovn_lb_vip *lb_vip,
+                         struct ovn_desired_flow_table *flow_table)
+{
+    uint64_t stub[1024 / 8];
+    struct ofpbuf ofpacts = OFPBUF_STUB_INITIALIZER(stub);
+
+    struct ofpact_conntrack *ct = ofpact_put_CT(&ofpacts);
+    ct->recirc_table = NX_CT_RECIRC_NONE;
+    ct->zone_src.field = mf_from_id(MFF_LOG_SNAT_ZONE);
+    ct->zone_src.ofs = 0;
+    ct->zone_src.n_bits = 16;
+    ct->flags = NX_CT_F_COMMIT;
+    ct->alg = 0;
+
+    size_t nat_offset;
+    nat_offset = ofpacts.size;
+    ofpbuf_pull(&ofpacts, nat_offset);
+
+    struct ofpact_nat *nat = ofpact_put_NAT(&ofpacts);
+    nat->flags = NX_NAT_F_SRC;
+    nat->range_af = AF_UNSPEC;
+
+    if (IN6_IS_ADDR_V4MAPPED(&lb_vip->vip)) {
+        nat->range_af = AF_INET;
+        nat->range.addr.ipv4.min = in6_addr_get_mapped_ipv4(&lb_vip->vip);
+    } else {
+        nat->range_af = AF_INET6;
+        nat->range.addr.ipv6.min = lb_vip->vip;
+    }
+    ofpacts.header = ofpbuf_push_uninit(&ofpacts, nat_offset);
+    ofpact_finish(&ofpacts, &ct->ofpact);
+
+    struct match match = MATCH_CATCHALL_INITIALIZER;
+    if (IN6_IS_ADDR_V4MAPPED(&lb_vip->vip)) {
+        match_set_dl_type(&match, htons(ETH_TYPE_IP));
+        match_set_ct_nw_dst(&match, nat->range.addr.ipv4.min);
+    } else {
+        match_set_dl_type(&match, htons(ETH_TYPE_IPV6));
+        match_set_ct_ipv6_dst(&match, &lb_vip->vip);
+    }
+
+    uint32_t ct_state = OVS_CS_F_TRACKED | OVS_CS_F_DST_NAT;
+    match_set_ct_state_masked(&match, ct_state, ct_state);
+
+    for (size_t i = 0; i < lb->slb->n_datapaths; i++) {
+        match_set_metadata(&match,
+                           htonll(lb->slb->datapaths[i]->tunnel_key));
+
+        ofctrl_add_flow(flow_table, OFTABLE_CT_SNAT_FOR_VIP, 100,
+                        lb->slb->header_.uuid.parts[0],
+                        &match, &ofpacts, &lb->slb->header_.uuid);
+    }
+
+    ofpbuf_uninit(&ofpacts);
+}
+
+static void
+consider_lb_hairpin_flows(const struct sbrec_load_balancer *sbrec_lb,
+                          const struct hmap *local_datapaths,
+                          struct ovn_desired_flow_table *flow_table)
+{
+    /* Check if we need to add flows or not.  If there is one datapath
+     * in the local_datapaths, it means all the datapaths of the lb
+     * will be in the local_datapaths. */
+    size_t i;
+    for (i = 0; i < sbrec_lb->n_datapaths; i++) {
+        if (get_local_datapath(local_datapaths,
+                               sbrec_lb->datapaths[i]->tunnel_key)) {
+            break;
+        }
+    }
+
+    if (i == sbrec_lb->n_datapaths) {
+        return;
+    }
+
+    struct ovn_controller_lb *lb = ovn_controller_lb_create(sbrec_lb);
+    uint8_t lb_proto = IPPROTO_TCP;
+    if (lb->slb->protocol && lb->slb->protocol[0]) {
+        if (!strcmp(lb->slb->protocol, "udp")) {
+            lb_proto = IPPROTO_UDP;
+        } else if (!strcmp(lb->slb->protocol, "sctp")) {
+            lb_proto = IPPROTO_SCTP;
+        }
+    }
+
+    for (i = 0; i < lb->n_vips; i++) {
+        struct ovn_lb_vip *lb_vip = &lb->vips[i];
+
+        for (size_t j = 0; j < lb_vip->n_backends; j++) {
+            struct ovn_lb_backend *lb_backend = &lb_vip->backends[j];
+
+            add_lb_vip_hairpin_flows(lb, lb_vip, lb_backend, lb_proto,
+                                     flow_table);
+        }
+
+        add_lb_ct_snat_vip_flows(lb, lb_vip, flow_table);
+    }
+
+    ovn_controller_lb_destroy(lb);
+}
+
+/* Adds OpenFlow flows to flow tables for each Load balancer VIPs and
+ * backends to handle the load balanced hairpin traffic. */
+static void
+add_lb_hairpin_flows(const struct sbrec_load_balancer_table *lb_table,
+                     const struct hmap *local_datapaths,
+                     struct ovn_desired_flow_table *flow_table)
+{
+    const struct sbrec_load_balancer *lb;
+    SBREC_LOAD_BALANCER_TABLE_FOR_EACH (lb, lb_table) {
+        consider_lb_hairpin_flows(lb, local_datapaths, flow_table);
+    }
+}
+
 /* Handles neighbor changes in mac_binding table. */
 void
 lflow_handle_changed_neighbors(
@@ -1203,6 +1429,8 @@ lflow_run(struct lflow_ctx_in *l_ctx_in, struct lflow_ctx_out *l_ctx_out)
     add_neighbor_flows(l_ctx_in->sbrec_port_binding_by_name,
                        l_ctx_in->mac_binding_table, l_ctx_in->local_datapaths,
                        l_ctx_out->flow_table);
+    add_lb_hairpin_flows(l_ctx_in->lb_table, l_ctx_in->local_datapaths,
+                         l_ctx_out->flow_table);
 }
 
 void
@@ -1248,20 +1476,61 @@ lflow_add_flows_for_datapath(const struct sbrec_datapath_binding *dp,
     const struct sbrec_logical_flow *lflow;
     SBREC_LOGICAL_FLOW_FOR_EACH_EQUAL (
         lflow, lf_row, l_ctx_in->sbrec_logical_flow_by_logical_datapath) {
-        if (!consider_logical_flow(lflow, &dhcp_opts, &dhcpv6_opts,
-                                   &nd_ra_opts, &controller_event_opts,
-                                   l_ctx_in, l_ctx_out)) {
+        if (!consider_logical_flow__(lflow, dp, &dhcp_opts, &dhcpv6_opts,
+                                     &nd_ra_opts, &controller_event_opts,
+                                     l_ctx_in, l_ctx_out)) {
             handled = false;
             l_ctx_out->conj_id_overflow = true;
-            break;
+            goto lflow_processing_end;
         }
     }
     sbrec_logical_flow_index_destroy_row(lf_row);
 
+    lf_row = sbrec_logical_flow_index_init_row(
+        l_ctx_in->sbrec_logical_flow_by_logical_dp_group);
+    /* There are far fewer datapath groups than logical flows. */
+    const struct sbrec_logical_dp_group *ldpg;
+    SBREC_LOGICAL_DP_GROUP_TABLE_FOR_EACH (ldpg,
+                                           l_ctx_in->logical_dp_group_table) {
+        bool found = false;
+        for (size_t i = 0; i < ldpg->n_datapaths; i++) {
+            if (ldpg->datapaths[i] == dp) {
+                found = true;
+                break;
+            }
+        }
+        if (!found) {
+            continue;
+        }
+
+        sbrec_logical_flow_index_set_logical_dp_group(lf_row, ldpg);
+        SBREC_LOGICAL_FLOW_FOR_EACH_EQUAL (
+            lflow, lf_row, l_ctx_in->sbrec_logical_flow_by_logical_dp_group) {
+            if (!consider_logical_flow__(lflow, dp, &dhcp_opts, &dhcpv6_opts,
+                                         &nd_ra_opts, &controller_event_opts,
+                                         l_ctx_in, l_ctx_out)) {
+                handled = false;
+                l_ctx_out->conj_id_overflow = true;
+                goto lflow_processing_end;
+            }
+        }
+    }
+lflow_processing_end:
+    sbrec_logical_flow_index_destroy_row(lf_row);
+
     dhcp_opts_destroy(&dhcp_opts);
     dhcp_opts_destroy(&dhcpv6_opts);
     nd_ra_opts_destroy(&nd_ra_opts);
     controller_event_opts_destroy(&controller_event_opts);
+
+    /* Add load balancer hairpin flows if the datapath has any load balancers
+     * associated. */
+    for (size_t i = 0; i < dp->n_load_balancers; i++) {
+        consider_lb_hairpin_flows(dp->load_balancers[i],
+                                  l_ctx_in->local_datapaths,
+                                  l_ctx_out->flow_table);
+    }
+
     return handled;
 }
 
@@ -1279,3 +1548,37 @@ lflow_handle_flows_for_lport(const struct sbrec_port_binding *pb,
     return lflow_handle_changed_ref(REF_TYPE_PORTBINDING, pb_ref_name,
                                     l_ctx_in, l_ctx_out, &changed);
 }
+
+bool
+lflow_handle_changed_lbs(struct lflow_ctx_in *l_ctx_in,
+                         struct lflow_ctx_out *l_ctx_out)
+{
+    const struct sbrec_load_balancer *lb;
+
+    SBREC_LOAD_BALANCER_TABLE_FOR_EACH_TRACKED (lb, l_ctx_in->lb_table) {
+        if (sbrec_load_balancer_is_deleted(lb)) {
+            VLOG_DBG("Remove hairpin flows for deleted load balancer "UUID_FMT,
+                     UUID_ARGS(&lb->header_.uuid));
+            ofctrl_remove_flows(l_ctx_out->flow_table, &lb->header_.uuid);
+        }
+    }
+
+    SBREC_LOAD_BALANCER_TABLE_FOR_EACH_TRACKED (lb, l_ctx_in->lb_table) {
+        if (sbrec_load_balancer_is_deleted(lb)) {
+            continue;
+        }
+
+        if (!sbrec_load_balancer_is_new(lb)) {
+            VLOG_DBG("Remove hairpin flows for updated load balancer "UUID_FMT,
+                     UUID_ARGS(&lb->header_.uuid));
+            ofctrl_remove_flows(l_ctx_out->flow_table, &lb->header_.uuid);
+        }
+
+        VLOG_DBG("Add load balancer hairpin flows for "UUID_FMT,
+                 UUID_ARGS(&lb->header_.uuid));
+        consider_lb_hairpin_flows(lb, l_ctx_in->local_datapaths,
+                                  l_ctx_out->flow_table);
+    }
+
+    return true;
+}
