@@ -188,6 +188,16 @@ struct sb_flow_ref {
  * relationship is 1 to N. A link is added when a flow addition is processed.
  * A link is removed when a flow deletion is processed, the desired flow
  * table is cleared, or the installed flow table is cleared.
+ *
+ * To ensure predictable behavior, the list of desired flows is maintained
+ * partially sorted in the following way (from least restrictive to most
+ * restrictive wrt. match):
+ * - allow flows without action conjunction.
+ * - drop flows without action conjunction.
+ * - a single flow with action conjunction.
+ *
+ * The first desired_flow in the list is the active one, the one that is
+ * actually installed.
  */
 struct installed_flow {
     struct ovn_flow flow;
@@ -199,13 +209,11 @@ struct installed_flow {
      * installed flow, e.g. when there are conflict/duplicated ACLs that
      * generates same match conditions). */
     struct ovs_list desired_refs;
-
-    /* The corresponding flow in desired table. It must be one of the flows in
-     * desired_refs list.  If there are more than one flows in references list,
-     * this is the one that is actually installed. */
-    struct desired_flow *desired_flow;
 };
 
+typedef bool
+(*desired_flow_match_cb)(const struct desired_flow *candidate,
+                         const void *arg);
 static struct desired_flow *desired_flow_alloc(
     uint8_t table_id,
     uint16_t priority,
@@ -213,15 +221,22 @@ static struct desired_flow *desired_flow_alloc(
     const struct match *match,
     const struct ofpbuf *actions);
 static struct desired_flow *desired_flow_lookup(
+    struct ovn_desired_flow_table *,
+    const struct ovn_flow *target);
+static struct desired_flow *desired_flow_lookup_check_uuid(
     struct ovn_desired_flow_table *,
     const struct ovn_flow *target,
-    const struct uuid *sb_uuid);
+    const struct uuid *);
+static struct desired_flow *desired_flow_lookup_conjunctive(
+    struct ovn_desired_flow_table *,
+    const struct ovn_flow *target);
 static void desired_flow_destroy(struct desired_flow *);
 
 static struct installed_flow *installed_flow_lookup(
     const struct ovn_flow *target);
 static void installed_flow_destroy(struct installed_flow *);
 static struct installed_flow *installed_flow_dup(struct desired_flow *);
+static struct desired_flow *installed_flow_get_active(struct installed_flow *);
 
 static uint32_t ovn_flow_match_hash(const struct ovn_flow *);
 static char *ovn_flow_to_string(const struct ovn_flow *);
@@ -787,35 +802,109 @@ ofctrl_recv(const struct ofp_header *oh, enum ofptype type)
         log_openflow_rl(&rl, VLL_DBG, oh, "OpenFlow packet ignored");
     }
 }
-
-static void
-link_installed_to_desired(struct installed_flow *i, struct desired_flow *d)
-{
-    if (i->desired_flow == d) {
-        return;
-    }
 
-    if (ovs_list_is_empty(&i->desired_refs)) {
-        ovs_assert(!i->desired_flow);
-        i->desired_flow = d;
-    }
-    ovs_list_insert(&i->desired_refs, &d->installed_ref_list_node);
-    d->installed_flow = i;
+static bool
+flow_action_has_drop(const struct ovn_flow *f)
+{
+    return f->ofpacts_len == 0;
 }
 
+static bool
+flow_action_has_conj(const struct ovn_flow *f)
+{
+    const struct ofpact *a = NULL;
+
+    OFPACT_FOR_EACH (a, f->ofpacts, f->ofpacts_len) {
+        if (a->type == OFPACT_CONJUNCTION) {
+            return true;
+        }
+    }
+    return false;
+}
+
+static bool
+flow_action_has_allow(const struct ovn_flow *f)
+{
+    return !flow_action_has_drop(f) && !flow_action_has_conj(f);
+}
+
+/* Returns true if flow 'a' is preferred over flow 'b'. */
+static bool
+flow_is_preferred(const struct ovn_flow *a, const struct ovn_flow *b)
+{
+    if (flow_action_has_allow(b)) {
+        return false;
+    }
+    if (flow_action_has_allow(a)) {
+        return true;
+    }
+    if (flow_action_has_drop(b)) {
+        return false;
+    }
+    if (flow_action_has_drop(a)) {
+        return true;
+    }
+
+    /* Flows 'a' and 'b' should never both have action conjunction. */
+    OVS_NOT_REACHED();
+}
+
+/* Adds the desired flow to the list of desired flows that have same match
+ * conditions as the installed flow.
+ *
+ * It is caller's responsibility to make sure the link between the pair didn't
+ * exist before.
+ *
+ * Returns true if the newly added desired flow is selected to be the active
+ * one.
+ */
+static bool
+link_installed_to_desired(struct installed_flow *i, struct desired_flow *d)
+{
+    struct desired_flow *f;
+
+    /* Find first 'f' such that 'd' is preferred over 'f'.  If no such desired
+     * flow exists then 'f' will point after the last element of the list.
+     */
+    LIST_FOR_EACH (f, installed_ref_list_node, &i->desired_refs) {
+        if (flow_is_preferred(&d->flow, &f->flow)) {
+            break;
+        }
+    }
+    ovs_list_insert(&f->installed_ref_list_node, &d->installed_ref_list_node);
+    d->installed_flow = i;
+    return installed_flow_get_active(i) == d;
+}
+
+/* Replaces 'old_desired' with 'new_desired' in the list of desired flows
+ * that have same match conditions as the installed flow.
+ */
 static void
+replace_installed_to_desired(struct installed_flow *i,
+                             struct desired_flow *old_desired,
+                             struct desired_flow *new_desired)
+{
+    ovs_assert(old_desired->installed_flow == i);
+    ovs_list_replace(&new_desired->installed_ref_list_node,
+                     &old_desired->installed_ref_list_node);
+    old_desired->installed_flow = NULL;
+    new_desired->installed_flow = i;
+}
+
+/* Removes the desired flow from the list of desired flows that have the same
+ * match conditions as the installed flow.
+ *
+ * Returns true if the desired flow was the previously active flow.
+ */
+static bool
 unlink_installed_to_desired(struct installed_flow *i, struct desired_flow *d)
 {
-    ovs_assert(i && i->desired_flow && !ovs_list_is_empty(&i->desired_refs));
+    struct desired_flow *old_active = installed_flow_get_active(i);
+
     ovs_assert(d && d->installed_flow == i);
     ovs_list_remove(&d->installed_ref_list_node);
     d->installed_flow = NULL;
-    if (i->desired_flow == d) {
-        i->desired_flow = ovs_list_is_empty(&i->desired_refs) ? NULL :
-            CONTAINER_OF(ovs_list_front(&i->desired_refs),
-                         struct desired_flow,
-                         installed_ref_list_node);
-    }
+    return old_active == d;
 }
 
 static void
@@ -939,7 +1028,7 @@ ofctrl_check_and_add_flow(struct ovn_desired_flow_table *flow_table,
     struct desired_flow *f = desired_flow_alloc(table_id, priority, cookie,
                                                 match, actions);
 
-    if (desired_flow_lookup(flow_table, &f->flow, sb_uuid)) {
+    if (desired_flow_lookup_check_uuid(flow_table, &f->flow, sb_uuid)) {
         if (log_duplicate_flow) {
             static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(5, 5);
             if (!VLOG_DROP_DBG(&rl)) {
@@ -979,14 +1068,15 @@ ofctrl_add_or_append_flow(struct ovn_desired_flow_table *desired_flows,
                           const struct ofpbuf *actions,
                           const struct uuid *sb_uuid)
 {
-    struct desired_flow *f = desired_flow_alloc(table_id, priority, cookie,
-                                                match, actions);
-
     struct desired_flow *existing;
-    existing = desired_flow_lookup(desired_flows, &f->flow, NULL);
+    struct desired_flow *f;
+
+    f = desired_flow_alloc(table_id, priority, cookie, match, actions);
+    existing = desired_flow_lookup_conjunctive(desired_flows, &f->flow);
     if (existing) {
-        /* There's already a flow with this particular match. Append the
-         * action to that flow rather than adding a new flow
+        /* There's already a flow with this particular match and action
+         * 'conjunction'. Append the action to that flow rather than
+         * adding a new flow.
          */
         uint64_t compound_stub[64 / 8];
         struct ofpbuf compound;
@@ -1214,7 +1304,6 @@ installed_flow_dup(struct desired_flow *src)
 {
     struct installed_flow *dst = xmalloc(sizeof *dst);
     ovs_list_init(&dst->desired_refs);
-    dst->desired_flow = NULL;
     dst->flow.table_id = src->flow.table_id;
     dst->flow.priority = src->flow.priority;
     minimatch_clone(&dst->flow.match, &src->flow.match);
@@ -1225,15 +1314,22 @@ installed_flow_dup(struct desired_flow *src)
     return dst;
 }
 
-/* Finds and returns a desired_flow in 'flow_table' whose key is identical to
- * 'target''s key, or NULL if there is none.
- *
- * If sb_uuid is not NULL, the function will also check if the found flow is
- * referenced by the sb_uuid. */
 static struct desired_flow *
-desired_flow_lookup(struct ovn_desired_flow_table *flow_table,
-                    const struct ovn_flow *target,
-                    const struct uuid *sb_uuid)
+installed_flow_get_active(struct installed_flow *f)
+{
+    if (!ovs_list_is_empty(&f->desired_refs)) {
+        return CONTAINER_OF(ovs_list_front(&f->desired_refs),
+                            struct desired_flow,
+                            installed_ref_list_node);
+    }
+    return NULL;
+}
+
+static struct desired_flow *
+desired_flow_lookup__(struct ovn_desired_flow_table *flow_table,
+                      const struct ovn_flow *target,
+                      desired_flow_match_cb match_cb,
+                      const void *arg)
 {
     struct desired_flow *d;
     HMAP_FOR_EACH_WITH_HASH (d, match_hmap_node, target->hash,
@@ -1242,20 +1338,76 @@ desired_flow_lookup(struct ovn_desired_flow_table *flow_table,
         if (f->table_id == target->table_id
             && f->priority == target->priority
             && minimatch_equal(&f->match, &target->match)) {
-            if (!sb_uuid) {
+
+            if (!match_cb || match_cb(d, arg)) {
                 return d;
             }
-            struct sb_flow_ref *sfr;
-            LIST_FOR_EACH (sfr, sb_list, &d->references) {
-                if (uuid_equals(sb_uuid, &sfr->sb_uuid)) {
-                    return d;
-                }
-            }
         }
     }
     return NULL;
 }
 
+/* Finds and returns a desired_flow in 'flow_table' whose key is identical to
+ * 'target''s key, or NULL if there is none.
+ */
+static struct desired_flow *
+desired_flow_lookup(struct ovn_desired_flow_table *flow_table,
+                    const struct ovn_flow *target)
+{
+    return desired_flow_lookup__(flow_table, target, NULL, NULL);
+}
+
+static bool
+flow_lookup_match_uuid_cb(const struct desired_flow *candidate,
+                          const void *arg)
+{
+    const struct uuid *sb_uuid = arg;
+    struct sb_flow_ref *sfr;
+
+    LIST_FOR_EACH (sfr, sb_list, &candidate->references) {
+        if (uuid_equals(sb_uuid, &sfr->sb_uuid)) {
+            return true;
+        }
+    }
+    return false;
+}
+
+/* Finds and returns a desired_flow in 'flow_table' whose key is identical to
+ * 'target''s key, or NULL if there is none.
+ *
+ * The function will also check if the found flow is referenced by the
+ * 'sb_uuid'.
+ */
+static struct desired_flow *
+desired_flow_lookup_check_uuid(struct ovn_desired_flow_table *flow_table,
+                            const struct ovn_flow *target,
+                            const struct uuid *sb_uuid)
+{
+    return desired_flow_lookup__(flow_table, target, flow_lookup_match_uuid_cb,
+                                 sb_uuid);
+}
+
+static bool
+flow_lookup_match_conj_cb(const struct desired_flow *candidate,
+                          const void *arg OVS_UNUSED)
+{
+    return flow_action_has_conj(&candidate->flow);
+}
+
+/* Finds and returns a desired_flow in 'flow_table' whose key is identical to
+ * 'target''s key, or NULL if there is none.
+ *
+ * The function will only return a matching flow if it contains action
+ * 'conjunction'.
+ */
+static struct desired_flow *
+desired_flow_lookup_conjunctive(struct ovn_desired_flow_table *flow_table,
+                                const struct ovn_flow *target)
+{
+    return desired_flow_lookup__(flow_table, target, flow_lookup_match_conj_cb,
+                                 NULL);
+}
+
 /* Finds and returns an installed_flow in installed_flows whose key is
  * identical to 'target''s key, or NULL if there is none. */
 static struct installed_flow *
@@ -1321,8 +1473,7 @@ static void
 installed_flow_destroy(struct installed_flow *f)
 {
     if (f) {
-        ovs_assert(ovs_list_is_empty(&f->desired_refs));
-        ovs_assert(!f->desired_flow);
+        ovs_assert(!installed_flow_get_active(f));
         ovn_flow_uninit(&f->flow);
         free(f);
     }
@@ -1524,6 +1675,7 @@ add_meter_string(struct ovn_extend_table_info *m_desired,
                                           &usable_protocols);
     if (!error) {
         add_meter_mod(&mm, msgs);
+        free(mm.meter.bands);
     } else {
         static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(5, 1);
         VLOG_ERR_RL(&rl, "new meter %s %s", error, meter_string);
@@ -1653,8 +1805,7 @@ update_installed_flows_by_compare(struct ovn_desired_flow_table *flow_table,
     struct installed_flow *i, *next;
     HMAP_FOR_EACH_SAFE (i, next, match_hmap_node, &installed_flows) {
         unlink_all_refs_for_installed_flow(i);
-        struct desired_flow *d =
-            desired_flow_lookup(flow_table, &i->flow, NULL);
+        struct desired_flow *d = desired_flow_lookup(flow_table, &i->flow);
         if (!d) {
             /* Installed flow is no longer desirable.  Delete it from the
              * switch and from installed_flows. */
@@ -1687,8 +1838,18 @@ update_installed_flows_by_compare(struct ovn_desired_flow_table *flow_table,
             /* Copy 'd' from 'flow_table' to installed_flows. */
             i = installed_flow_dup(d);
             hmap_insert(&installed_flows, &i->match_hmap_node, i->flow.hash);
+            link_installed_to_desired(i, d);
+        } else if (!d->installed_flow) {
+            /* This is a desired_flow that conflicts with one installed
+             * previously but not linked yet.  However, if this flow becomes
+             * active, e.g., it is less restrictive than the previous active
+             * flow then modify the installed flow.
+             */
+            if (link_installed_to_desired(i, d)) {
+                installed_flow_mod(&i->flow, &d->flow, msgs);
+                ovn_flow_log(&i->flow, "updating installed (conflict)");
+            }
         }
-        link_installed_to_desired(i, d);
     }
 }
 
@@ -1740,11 +1901,10 @@ merge_tracked_flows(struct ovn_desired_flow_table *flow_table)
             /* del_f must have been installed, otherwise it should have been
              * removed during track_flow_add_or_modify. */
             ovs_assert(del_f->installed_flow);
+
             if (!f->installed_flow) {
                 /* f is not installed yet. */
-                struct installed_flow *i = del_f->installed_flow;
-                unlink_installed_to_desired(i, del_f);
-                link_installed_to_desired(i, f);
+                replace_installed_to_desired(del_f->installed_flow, del_f, f);
             } else {
                 /* f has been installed before, and now was updated to exact
                  * the same flow as del_f. */
@@ -1778,22 +1938,22 @@ update_installed_flows_by_track(struct ovn_desired_flow_table *flow_table,
             /* The desired flow was deleted */
             if (f->installed_flow) {
                 struct installed_flow *i = f->installed_flow;
-                unlink_installed_to_desired(i, f);
+                bool was_active = unlink_installed_to_desired(i, f);
+                struct desired_flow *d = installed_flow_get_active(i);
 
-                if (!i->desired_flow) {
+                if (!d) {
                     installed_flow_del(&i->flow, msgs);
                     ovn_flow_log(&i->flow, "removing installed (tracked)");
 
                     hmap_remove(&installed_flows, &i->match_hmap_node);
                     installed_flow_destroy(i);
-                } else {
+                } else if (was_active) {
                     /* There are other desired flow(s) referencing this
                      * installed flow, so update the OVS flow for the new
                      * active flow (at least the cookie will be different,
                      * even if the actions are the same). */
-                    struct desired_flow *d = i->desired_flow;
-                    ovn_flow_log(&i->flow, "updating installed (tracked)");
                     installed_flow_mod(&i->flow, &d->flow, msgs);
+                    ovn_flow_log(&i->flow, "updating installed (tracked)");
                 }
             }
             desired_flow_destroy(f);
@@ -1810,15 +1970,22 @@ update_installed_flows_by_track(struct ovn_desired_flow_table *flow_table,
                 hmap_insert(&installed_flows, &new_node->match_hmap_node,
                             new_node->flow.hash);
                 link_installed_to_desired(new_node, f);
-            } else if (i->desired_flow == f) {
+            } else if (installed_flow_get_active(i) == f) {
                 /* The installed flow is installed for f, but f has change
                  * tracked, so it must have been modified. */
-                ovn_flow_log(&i->flow, "updating installed (tracked)");
                 installed_flow_mod(&i->flow, &f->flow, msgs);
+                ovn_flow_log(&i->flow, "updating installed (tracked)");
             } else {
                 /* Adding a new flow that conflicts with an existing installed
-                 * flow, so just add it to the link. */
-                link_installed_to_desired(i, f);
+                 * flow, so add it to the link.  If this flow becomes active,
+                 * e.g., it is less restrictive than the previous active flow
+                 * then modify the installed flow.
+                 */
+                if (link_installed_to_desired(i, f)) {
+                    installed_flow_mod(&i->flow, &f->flow, msgs);
+                    ovn_flow_log(&i->flow,
+                                 "updating installed (tracked conflict)");
+                }
             }
             /* The track_list_node emptyness is used to check if the node is
              * already added to track list, so initialize it again here. */
@@ -1865,19 +2032,23 @@ ofctrl_put(struct ovn_desired_flow_table *flow_table,
     bool need_put = false;
     if (flow_changed || skipped_last_time || need_reinstall_flows) {
         need_put = true;
+        old_nb_cfg = nb_cfg;
     } else if (nb_cfg != old_nb_cfg) {
         /* nb_cfg changed since last ofctrl_put() call */
         if (cur_cfg == old_nb_cfg) {
-            /* we were up-to-date already, so just update with the
-             * new nb_cfg */
-            cur_cfg = nb_cfg;
+            /* If there are no updates pending, we were up-to-date already,
+             * update with the new nb_cfg.
+             */
+            if (ovs_list_is_empty(&flow_updates)) {
+                cur_cfg = nb_cfg;
+                old_nb_cfg = nb_cfg;
+            }
         } else {
             need_put = true;
+            old_nb_cfg = nb_cfg;
         }
     }
 
-    old_nb_cfg = nb_cfg;
-
     if (!need_put) {
         VLOG_DBG("ofctrl_put not needed");
         return;
