@@ -93,6 +93,20 @@ struct gen_opts_map {
 #define DHCP_OPT_DOMAIN_SEARCH_LIST \
     DHCP_OPTION("domain_search_list", 119, "domains")
 
+#define DHCP_OPT_BROADCAST_ADDRESS \
+    DHCP_OPTION("broadcast_address", 28, "ipv4")
+
+#define DHCP_OPT_BOOTFILE_CODE 67
+
+/* Use unused 254 option for iPXE bootfile_name_alt userdata DHCP option.
+ * This option code is replaced by 67 when sending the DHCP reply.
+ */
+#define DHCP_OPT_BOOTFILE_ALT_CODE 254
+#define DHCP_OPT_BOOTFILE_ALT DHCP_OPTION("bootfile_name_alt", \
+                                          DHCP_OPT_BOOTFILE_ALT_CODE, "str")
+
+#define DHCP_OPT_ETHERBOOT	175
+
 #define DHCP_OPT_ARP_CACHE_TIMEOUT \
     DHCP_OPTION("arp_cache_timeout", 35, "uint32")
 #define DHCP_OPT_TCP_KEEPALIVE_INTERVAL \
@@ -189,6 +203,9 @@ struct dhcp_opt6_header {
 #define DHCPV6_MSG_TYPE_ADVT        2
 #define DHCPV6_MSG_TYPE_REQUEST     3
 #define DHCPV6_MSG_TYPE_CONFIRM     4
+#define DHCPV6_MSG_TYPE_RENEW       5
+#define DHCPV6_MSG_TYPE_REBIND      6
+
 #define DHCPV6_MSG_TYPE_REPLY       7
 #define DHCPV6_MSG_TYPE_DECLINE     9
 #define DHCPV6_MSG_TYPE_INFO_REQ    11
@@ -406,23 +423,17 @@ controller_event_opts_destroy(struct controller_event_options *opts)
     }
 }
 
-static inline bool
-ipv6_addr_is_routable_multicast(const struct in6_addr *ip) {
-    if (!ipv6_addr_is_multicast(ip)) {
-        return false;
-    }
+bool ipv6_addr_is_routable_multicast(const struct in6_addr *);
 
-    /* Check multicast group scope, RFC 4291, 2.7. */
-    switch (ip->s6_addr[1] & 0x0F) {
-    case 0x00:
-    case 0x01:
-    case 0x02:
-    case 0x03:
-    case 0x0F:
-        return false;
-    default:
-        return true;
-    }
+static inline bool
+ipv6_addr_is_host_zero(const struct in6_addr *prefix,
+                       const struct in6_addr *mask)
+{
+    /* host-bits-non-zero <=> (prefix ^ mask) & prefix. */
+    struct in6_addr tmp = ipv6_addr_bitxor(prefix, mask);
+
+    tmp = ipv6_addr_bitand(&tmp, prefix);
+    return ipv6_is_zero(&tmp);
 }
 
 #define IPV6_EXT_HEADER_LEN 8
