@@ -13,17 +13,22 @@
  */
 
 #include <config.h>
+
+#include "ovn-util.h"
+
+#include <ctype.h>
 #include <unistd.h>
 
 #include "daemon.h"
-#include "ovn-util.h"
-#include "ovn-dirs.h"
-#include "openvswitch/vlog.h"
+#include "include/ovn/actions.h"
 #include "openvswitch/ofp-parse.h"
+#include "openvswitch/vlog.h"
+#include "ovn-dirs.h"
 #include "ovn-nb-idl.h"
 #include "ovn-sb-idl.h"
+#include "socket-util.h"
+#include "svec.h"
 #include "unixctl.h"
-#include <ctype.h>
 
 VLOG_DEFINE_THIS_MODULE(ovn_util);
 
@@ -240,27 +245,37 @@ extract_ip_addresses(const char *address, struct lport_addresses *laddrs)
 bool
 extract_lrp_networks(const struct nbrec_logical_router_port *lrp,
                      struct lport_addresses *laddrs)
+{
+    return extract_lrp_networks__(lrp->mac, lrp->networks, lrp->n_networks,
+                                  laddrs);
+}
+
+/* Separate out the body of 'extract_lrp_networks()' for use from DDlog,
+ * which does not know the 'nbrec_logical_router_port' type. */
+bool
+extract_lrp_networks__(char *mac, char **networks, size_t n_networks,
+                       struct lport_addresses *laddrs)
 {
     memset(laddrs, 0, sizeof *laddrs);
 
-    if (!eth_addr_from_string(lrp->mac, &laddrs->ea)) {
+    if (!eth_addr_from_string(mac, &laddrs->ea)) {
         laddrs->ea = eth_addr_zero;
         return false;
     }
     snprintf(laddrs->ea_s, sizeof laddrs->ea_s, ETH_ADDR_FMT,
              ETH_ADDR_ARGS(laddrs->ea));
 
-    for (int i = 0; i < lrp->n_networks; i++) {
+    for (int i = 0; i < n_networks; i++) {
         ovs_be32 ip4;
         struct in6_addr ip6;
         unsigned int plen;
         char *error;
 
-        error = ip_parse_cidr(lrp->networks[i], &ip4, &plen);
+        error = ip_parse_cidr(networks[i], &ip4, &plen);
         if (!error) {
             if (!ip4) {
                 static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(5, 1);
-                VLOG_WARN_RL(&rl, "bad 'networks' %s", lrp->networks[i]);
+                VLOG_WARN_RL(&rl, "bad 'networks' %s", networks[i]);
                 continue;
             }
 
@@ -269,13 +284,13 @@ extract_lrp_networks(const struct nbrec_logical_router_port *lrp,
         }
         free(error);
 
-        error = ipv6_parse_cidr(lrp->networks[i], &ip6, &plen);
+        error = ipv6_parse_cidr(networks[i], &ip6, &plen);
         if (!error) {
             add_ipv6_netaddr(laddrs, ip6, plen);
         } else {
             static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(1, 1);
             VLOG_INFO_RL(&rl, "invalid syntax '%s' in networks",
-                         lrp->networks[i]);
+                         networks[i]);
             free(error);
         }
     }
@@ -320,6 +335,12 @@ extract_sbrec_binding_first_mac(const struct sbrec_port_binding *binding,
     return ret;
 }
 
+bool
+lport_addresses_is_empty(struct lport_addresses *laddrs)
+{
+    return !laddrs->n_ipv4_addrs && !laddrs->n_ipv6_addrs;
+}
+
 void
 destroy_lport_addresses(struct lport_addresses *laddrs)
 {
@@ -327,6 +348,23 @@ destroy_lport_addresses(struct lport_addresses *laddrs)
     free(laddrs->ipv6_addrs);
 }
 
+/* Go through 'addresses' and add found IPv4 addresses to 'ipv4_addrs' and
+ * IPv6 addresses to 'ipv6_addrs'. */
+void
+split_addresses(const char *addresses, struct svec *ipv4_addrs,
+                struct svec *ipv6_addrs)
+{
+    struct lport_addresses laddrs;
+    extract_lsp_addresses(addresses, &laddrs);
+    for (size_t k = 0; k < laddrs.n_ipv4_addrs; k++) {
+        svec_add(ipv4_addrs, laddrs.ipv4_addrs[k].addr_s);
+    }
+    for (size_t k = 0; k < laddrs.n_ipv6_addrs; k++) {
+        svec_add(ipv6_addrs, laddrs.ipv6_addrs[k].addr_s);
+    }
+    destroy_lport_addresses(&laddrs);
+}
+
 /* Allocates a key for NAT conntrack zone allocation for a provided
  * 'key' record and a 'type'.
  *
@@ -468,33 +506,43 @@ uint32_t
 sbrec_logical_flow_hash(const struct sbrec_logical_flow *lf)
 {
     const struct sbrec_datapath_binding *ld = lf->logical_datapath;
-    if (!ld) {
-        return 0;
-    }
+    uint32_t hash = ovn_logical_flow_hash(lf->table_id, lf->pipeline,
+                                          lf->priority, lf->match,
+                                          lf->actions);
 
-    return ovn_logical_flow_hash(&ld->header_.uuid,
-                                 lf->table_id, lf->pipeline,
-                                 lf->priority, lf->match, lf->actions);
+    return ld ? ovn_logical_flow_hash_datapath(&ld->header_.uuid, hash) : hash;
 }
 
 uint32_t
-ovn_logical_flow_hash(const struct uuid *logical_datapath,
-                      uint8_t table_id, const char *pipeline,
+ovn_logical_flow_hash(uint8_t table_id, const char *pipeline,
                       uint16_t priority,
                       const char *match, const char *actions)
 {
-    size_t hash = uuid_hash(logical_datapath);
-    hash = hash_2words((table_id << 16) | priority, hash);
+    size_t hash = hash_2words((table_id << 16) | priority, 0);
     hash = hash_string(pipeline, hash);
     hash = hash_string(match, hash);
     return hash_string(actions, hash);
 }
 
+uint32_t
+ovn_logical_flow_hash_datapath(const struct uuid *logical_datapath,
+                               uint32_t hash)
+{
+    return hash_add(hash, uuid_hash(logical_datapath));
+}
+
 bool
 datapath_is_switch(const struct sbrec_datapath_binding *ldp)
 {
     return smap_get(&ldp->external_ids, "logical-switch") != NULL;
 }
+
+int
+datapath_snat_ct_zone(const struct sbrec_datapath_binding *dp)
+{
+    return smap_get_int(&dp->external_ids, "snat-ct-zone", -1);
+}
+
 
 struct tnlid_node {
     struct hmap_node hmap_node;
@@ -511,24 +559,21 @@ ovn_destroy_tnlids(struct hmap *tnlids)
     hmap_destroy(tnlids);
 }
 
-void
+bool
 ovn_add_tnlid(struct hmap *set, uint32_t tnlid)
 {
-    struct tnlid_node *node = xmalloc(sizeof *node);
-    hmap_insert(set, &node->hmap_node, hash_int(tnlid, 0));
-    node->tnlid = tnlid;
-}
-
-bool
-ovn_tnlid_in_use(const struct hmap *set, uint32_t tnlid)
-{
-    const struct tnlid_node *node;
-    HMAP_FOR_EACH_IN_BUCKET (node, hmap_node, hash_int(tnlid, 0), set) {
+    uint32_t hash = hash_int(tnlid, 0);
+    struct tnlid_node *node;
+    HMAP_FOR_EACH_IN_BUCKET (node, hmap_node, hash, set) {
         if (node->tnlid == tnlid) {
-            return true;
+            return false;
         }
     }
-    return false;
+
+    node = xmalloc(sizeof *node);
+    hmap_insert(set, &node->hmap_node, hash);
+    node->tnlid = tnlid;
+    return true;
 }
 
 static uint32_t
@@ -543,8 +588,7 @@ ovn_allocate_tnlid(struct hmap *set, const char *name, uint32_t min,
 {
     for (uint32_t tnlid = next_tnlid(*hint, min, max); tnlid != *hint;
          tnlid = next_tnlid(tnlid, min, max)) {
-        if (!ovn_tnlid_in_use(set, tnlid)) {
-            ovn_add_tnlid(set, tnlid);
+        if (ovn_add_tnlid(set, tnlid)) {
             *hint = tnlid;
             return tnlid;
         }
@@ -562,33 +606,24 @@ ovn_chassis_redirect_name(const char *port_name)
 }
 
 bool
-ip46_parse_cidr(const char *str, struct v46_ip *prefix, unsigned int *plen)
+ip46_parse_cidr(const char *str, struct in6_addr *prefix, unsigned int *plen)
 {
-    memset(prefix, 0, sizeof *prefix);
+    ovs_be32 ipv4;
+    char *error = ip_parse_cidr(str, &ipv4, plen);
 
-    char *error = ip_parse_cidr(str, &prefix->ipv4, plen);
     if (!error) {
-        prefix->family = AF_INET;
+        in6_addr_set_mapped_ipv4(prefix, ipv4);
         return true;
     }
     free(error);
-    error = ipv6_parse_cidr(str, &prefix->ipv6, plen);
+    error = ipv6_parse_cidr(str, prefix, plen);
     if (!error) {
-        prefix->family = AF_INET6;
         return true;
     }
     free(error);
     return false;
 }
 
-bool
-ip46_equals(const struct v46_ip *addr1, const struct v46_ip *addr2)
-{
-    return (addr1->family == addr2->family &&
-            (addr1->family == AF_INET ? addr1->ipv4 == addr2->ipv4 :
-             IN6_ARE_ADDR_EQUAL(&addr1->ipv6, &addr2->ipv6)));
-}
-
 /* The caller must free the returned string. */
 char *
 normalize_ipv4_prefix(ovs_be32 ipv4, unsigned int plen)
@@ -603,12 +638,12 @@ normalize_ipv4_prefix(ovs_be32 ipv4, unsigned int plen)
 
 /* The caller must free the returned string. */
 char *
-normalize_ipv6_prefix(struct in6_addr ipv6, unsigned int plen)
+normalize_ipv6_prefix(const struct in6_addr *ipv6, unsigned int plen)
 {
     char network_s[INET6_ADDRSTRLEN];
 
     struct in6_addr mask = ipv6_create_mask(plen);
-    struct in6_addr network = ipv6_addr_bitand(&ipv6, &mask);
+    struct in6_addr network = ipv6_addr_bitand(ipv6, &mask);
 
     inet_ntop(AF_INET6, &network, network_s, INET6_ADDRSTRLEN);
     if (plen == 128) {
@@ -619,12 +654,12 @@ normalize_ipv6_prefix(struct in6_addr ipv6, unsigned int plen)
 }
 
 char *
-normalize_v46_prefix(const struct v46_ip *prefix, unsigned int plen)
+normalize_v46_prefix(const struct in6_addr *prefix, unsigned int plen)
 {
-    if (prefix->family == AF_INET) {
-        return normalize_ipv4_prefix(prefix->ipv4, plen);
+    if (IN6_IS_ADDR_V4MAPPED(prefix)) {
+        return normalize_ipv4_prefix(in6_addr_get_mapped_ipv4(prefix), plen);
     } else {
-        return normalize_ipv6_prefix(prefix->ipv6, plen);
+        return normalize_ipv6_prefix(prefix, plen);
     }
 }
 
@@ -661,3 +696,44 @@ ovn_smap_get_uint(const struct smap *smap, const char *key, unsigned int def)
 
     return u_value;
 }
+
+/* For a 'key' of the form "IP:port" or just "IP", sets 'port' and
+ * 'ip_address'.  The caller must free() the memory allocated for
+ * 'ip_address'.
+ * Returns true if parsing of 'key' was successful, false otherwise.
+ */
+bool
+ip_address_and_port_from_lb_key(const char *key, char **ip_address,
+                                uint16_t *port, int *addr_family)
+{
+    struct sockaddr_storage ss;
+    if (!inet_parse_active(key, 0, &ss, false)) {
+        static struct vlog_rate_limit rl = VLOG_RATE_LIMIT_INIT(5, 1);
+        VLOG_WARN_RL(&rl, "bad ip address or port for load balancer key %s",
+                     key);
+        *ip_address = NULL;
+        *port = 0;
+        *addr_family = 0;
+        return false;
+    }
+
+    struct ds s = DS_EMPTY_INITIALIZER;
+    ss_format_address_nobracks(&ss, &s);
+    *ip_address = ds_steal_cstr(&s);
+    *port = ss_get_port(&ss);
+    *addr_family = ss.ss_family;
+    return true;
+}
+
+/* Increment this for any logical flow changes or if existing OVN action is
+ * modified. */
+#define OVN_INTERNAL_MINOR_VER 0
+
+/* Returns the OVN version. The caller must free the returned value. */
+char *
+ovn_get_internal_version(void)
+{
+    return xasprintf("%s-%s-%d.%d", OVN_PACKAGE_VERSION,
+                     sbrec_get_db_version(),
+                     N_OVNACTS, OVN_INTERNAL_MINOR_VER);
+}
