@@ -109,13 +109,14 @@ main(int argc, char *argv[])
     sbctl_cmd_init();
 
     /* Check if options are set via env var. */
-    argv = ovs_cmdl_env_parse_all(&argc, argv, getenv("OVN_SBCTL_OPTIONS"));
+    char **argv_ = ovs_cmdl_env_parse_all(&argc, argv,
+                                          getenv("OVN_SBCTL_OPTIONS"));
 
     /* Parse command line. */
-    char *args = process_escape_args(argv);
+    char *args = process_escape_args(argv_);
     shash_init(&local_options);
-    parse_options(argc, argv, &local_options);
-    char *error = ctl_parse_commands(argc - optind, argv + optind,
+    parse_options(argc, argv_, &local_options);
+    char *error = ctl_parse_commands(argc - optind, argv_ + optind,
                                      &local_options, &commands, &n_commands);
     if (error) {
         ctl_fatal("%s", error);
@@ -149,8 +150,7 @@ main(int argc, char *argv[])
         if (seqno != ovsdb_idl_get_seqno(idl)) {
             seqno = ovsdb_idl_get_seqno(idl);
             if (do_sbctl(args, commands, n_commands, idl)) {
-                free(args);
-                exit(EXIT_SUCCESS);
+                break;
             }
         }
 
@@ -159,6 +159,13 @@ main(int argc, char *argv[])
             poll_block();
         }
     }
+
+    for (int i = 0; i < argc; i++) {
+        free(argv_[i]);
+    }
+    free(argv_);
+    free(args);
+    exit(EXIT_SUCCESS);
 }
 
 static void
@@ -521,6 +528,7 @@ pre_get_info(struct ctl_context *ctx)
     ovsdb_idl_add_column(ctx->idl, &sbrec_port_binding_col_datapath);
 
     ovsdb_idl_add_column(ctx->idl, &sbrec_logical_flow_col_logical_datapath);
+    ovsdb_idl_add_column(ctx->idl, &sbrec_logical_flow_col_logical_dp_group);
     ovsdb_idl_add_column(ctx->idl, &sbrec_logical_flow_col_pipeline);
     ovsdb_idl_add_column(ctx->idl, &sbrec_logical_flow_col_actions);
     ovsdb_idl_add_column(ctx->idl, &sbrec_logical_flow_col_priority);
@@ -528,6 +536,8 @@ pre_get_info(struct ctl_context *ctx)
     ovsdb_idl_add_column(ctx->idl, &sbrec_logical_flow_col_match);
     ovsdb_idl_add_column(ctx->idl, &sbrec_logical_flow_col_external_ids);
 
+    ovsdb_idl_add_column(ctx->idl, &sbrec_logical_dp_group_col_datapaths);
+
     ovsdb_idl_add_column(ctx->idl, &sbrec_datapath_binding_col_external_ids);
 
     ovsdb_idl_add_column(ctx->idl, &sbrec_ip_multicast_col_datapath);
@@ -542,6 +552,11 @@ pre_get_info(struct ctl_context *ctx)
     ovsdb_idl_add_column(ctx->idl, &sbrec_mac_binding_col_logical_port);
     ovsdb_idl_add_column(ctx->idl, &sbrec_mac_binding_col_ip);
     ovsdb_idl_add_column(ctx->idl, &sbrec_mac_binding_col_mac);
+
+    ovsdb_idl_add_column(ctx->idl, &sbrec_load_balancer_col_datapaths);
+    ovsdb_idl_add_column(ctx->idl, &sbrec_load_balancer_col_vips);
+    ovsdb_idl_add_column(ctx->idl, &sbrec_load_balancer_col_name);
+    ovsdb_idl_add_column(ctx->idl, &sbrec_load_balancer_col_protocol);
 }
 
 static struct cmd_show_table cmd_show_tables[] = {
@@ -705,38 +720,43 @@ pipeline_encode(const char *pl)
     OVS_NOT_REACHED();
 }
 
+struct sbctl_lflow {
+    const struct sbrec_logical_flow *lflow;
+    const struct sbrec_datapath_binding *dp;
+};
+
 static int
-lflow_cmp(const void *lf1_, const void *lf2_)
+sbctl_lflow_cmp(const void *a_, const void *b_)
 {
-    const struct sbrec_logical_flow *const *lf1p = lf1_;
-    const struct sbrec_logical_flow *const *lf2p = lf2_;
-    const struct sbrec_logical_flow *lf1 = *lf1p;
-    const struct sbrec_logical_flow *lf2 = *lf2p;
+    const struct sbctl_lflow *a_ctl_lflow = a_;
+    const struct sbctl_lflow *b_ctl_lflow = b_;
 
-    int pl1 = pipeline_encode(lf1->pipeline);
-    int pl2 = pipeline_encode(lf2->pipeline);
+    const struct sbrec_logical_flow *a = a_ctl_lflow->lflow;
+    const struct sbrec_logical_flow *b = b_ctl_lflow->lflow;
 
-#define CMP(expr) \
-    do { \
-        int res; \
-        res = (expr); \
-        if (res) { \
-            return res; \
-        } \
-    } while (0)
+    const struct sbrec_datapath_binding *adb = a_ctl_lflow->dp;
+    const struct sbrec_datapath_binding *bdb = b_ctl_lflow->dp;
+    const char *a_name = smap_get_def(&adb->external_ids, "name", "");
+    const char *b_name = smap_get_def(&bdb->external_ids, "name", "");
+    int cmp = strcmp(a_name, b_name);
+    if (cmp) {
+        return cmp;
+    }
 
-    CMP(uuid_compare_3way(&lf1->logical_datapath->header_.uuid,
-                          &lf2->logical_datapath->header_.uuid));
-    CMP(pl1 - pl2);
-    CMP(lf1->table_id > lf2->table_id ? 1 :
-            (lf1->table_id < lf2->table_id ? -1 : 0));
-    CMP(lf1->priority > lf2->priority ? -1 :
-            (lf1->priority < lf2->priority ? 1 : 0));
-    CMP(strcmp(lf1->match, lf2->match));
+    cmp = uuid_compare_3way(&adb->header_.uuid, &bdb->header_.uuid);
+    if (cmp) {
+        return cmp;
+    }
 
-#undef CMP
-
-    return 0;
+    int a_pipeline = pipeline_encode(a->pipeline);
+    int b_pipeline = pipeline_encode(b->pipeline);
+    return (a_pipeline > b_pipeline ? 1
+            : a_pipeline < b_pipeline ? -1
+            : a->table_id > b->table_id ? 1
+            : a->table_id < b->table_id ? -1
+            : a->priority > b->priority ? -1
+            : a->priority < b->priority ? 1
+            : strcmp(a->match, b->match));
 }
 
 static char *
@@ -1010,6 +1030,85 @@ cmd_lflow_list_chassis(struct ctl_context *ctx, struct vconn *vconn,
     }
 }
 
+static void
+cmd_lflow_list_load_balancers(struct ctl_context *ctx, struct vconn *vconn,
+                              const struct sbrec_datapath_binding *datapath,
+                              bool stats, bool print_uuid)
+{
+    const struct sbrec_load_balancer *lb;
+    const struct sbrec_load_balancer *lb_prev = NULL;
+    SBREC_LOAD_BALANCER_FOR_EACH (lb, ctx->idl) {
+        bool dp_found = false;
+        if (datapath) {
+            size_t i;
+            for (i = 0; i < lb->n_datapaths; i++) {
+                if (datapath == lb->datapaths[i]) {
+                    dp_found = true;
+                    break;
+                }
+            }
+            if (!dp_found) {
+                continue;
+            }
+        }
+
+        if (!lb_prev) {
+            printf("\nLoad Balancers:\n");
+        }
+
+        printf("  ");
+        print_uuid_part(&lb->header_.uuid, print_uuid);
+        printf("name=\"%s\", protocol=\"%s\", ", lb->name, lb->protocol);
+        if (!dp_found) {
+            for (size_t i = 0; i < lb->n_datapaths; i++) {
+                print_vflow_datapath_name(lb->datapaths[i], true);
+            }
+        }
+
+        printf("\n  vips:\n");
+        struct smap_node *node;
+        SMAP_FOR_EACH (node, &lb->vips) {
+            printf("    %s = %s\n", node->key, node->value);
+        }
+        printf("\n");
+
+        if (vconn) {
+            sbctl_dump_openflow(vconn, &lb->header_.uuid, stats);
+        }
+
+        lb_prev = lb;
+    }
+}
+
+static bool
+datapath_group_contains_datapath(const struct sbrec_logical_dp_group *g,
+                                 const struct sbrec_datapath_binding *dp)
+{
+    if (!g || !dp) {
+        return false;
+    }
+    for (size_t i = 0; i < g->n_datapaths; i++) {
+        if (g->datapaths[i] == dp) {
+            return true;
+        }
+    }
+    return false;
+}
+
+static void
+sbctl_lflow_add(struct sbctl_lflow **lflows,
+                size_t *n_flows, size_t *n_capacity,
+                const struct sbrec_logical_flow *lflow,
+                const struct sbrec_datapath_binding *dp)
+{
+    if (*n_flows == *n_capacity) {
+        *lflows = x2nrealloc(*lflows, n_capacity, sizeof **lflows);
+    }
+    (*lflows)[*n_flows].lflow = lflow;
+    (*lflows)[*n_flows].dp = dp;
+    (*n_flows)++;
+}
+
 static void
 cmd_lflow_list(struct ctl_context *ctx)
 {
@@ -1041,31 +1140,42 @@ cmd_lflow_list(struct ctl_context *ctx)
     struct vconn *vconn = sbctl_open_vconn(&ctx->options);
     bool stats = shash_find(&ctx->options, "--stats") != NULL;
 
-    const struct sbrec_logical_flow **lflows = NULL;
+    struct sbctl_lflow *lflows = NULL;
     size_t n_flows = 0;
     size_t n_capacity = 0;
     const struct sbrec_logical_flow *lflow;
+    const struct sbrec_logical_dp_group *dp_group;
     SBREC_LOGICAL_FLOW_FOR_EACH (lflow, ctx->idl) {
-        if (datapath && lflow->logical_datapath != datapath) {
+        if (datapath
+            && lflow->logical_datapath != datapath
+            && !datapath_group_contains_datapath(lflow->logical_dp_group,
+                                                 datapath)) {
             continue;
         }
-
-        if (n_flows == n_capacity) {
-            lflows = x2nrealloc(lflows, &n_capacity, sizeof *lflows);
+        if (datapath) {
+            sbctl_lflow_add(&lflows, &n_flows, &n_capacity, lflow, datapath);
+            continue;
+        }
+        if (lflow->logical_datapath) {
+            sbctl_lflow_add(&lflows, &n_flows, &n_capacity,
+                            lflow, lflow->logical_datapath);
+        }
+        dp_group = lflow->logical_dp_group;
+        for (size_t i = 0; dp_group && i < dp_group->n_datapaths; i++) {
+            sbctl_lflow_add(&lflows, &n_flows, &n_capacity,
+                            lflow, dp_group->datapaths[i]);
         }
-        lflows[n_flows] = lflow;
-        n_flows++;
     }
 
     if (n_flows) {
-        qsort(lflows, n_flows, sizeof *lflows, lflow_cmp);
+        qsort(lflows, n_flows, sizeof *lflows, sbctl_lflow_cmp);
     }
 
     bool print_uuid = shash_find(&ctx->options, "--uuid") != NULL;
 
-    const struct sbrec_logical_flow *prev = NULL;
+    const struct sbctl_lflow *curr, *prev = NULL;
     for (size_t i = 0; i < n_flows; i++) {
-        lflow = lflows[i];
+        curr = &lflows[i];
 
         /* Figure out whether to print this particular flow.  By default, we
          * print all flows, but if any UUIDs were listed on the command line
@@ -1074,7 +1184,7 @@ cmd_lflow_list(struct ctl_context *ctx)
         if (ctx->argc > 1) {
             include = false;
             for (size_t j = 1; j < ctx->argc; j++) {
-                if (is_partial_uuid_match(&lflow->header_.uuid,
+                if (is_partial_uuid_match(&curr->lflow->header_.uuid,
                                           ctx->argv[j])) {
                     include = true;
                     break;
@@ -1090,27 +1200,28 @@ cmd_lflow_list(struct ctl_context *ctx)
         /* Print a header line for this datapath or pipeline, if we haven't
          * already done so. */
         if (!prev
-            || prev->logical_datapath != lflow->logical_datapath
-            || strcmp(prev->pipeline, lflow->pipeline)) {
+            || prev->dp != curr->dp
+            || strcmp(prev->lflow->pipeline, curr->lflow->pipeline)) {
             printf("Datapath: ");
-            print_datapath_name(lflow->logical_datapath);
+            print_datapath_name(curr->dp);
             printf(" ("UUID_FMT")  Pipeline: %s\n",
-                   UUID_ARGS(&lflow->logical_datapath->header_.uuid),
-                   lflow->pipeline);
+                   UUID_ARGS(&curr->dp->header_.uuid),
+                   curr->lflow->pipeline);
         }
 
         /* Print the flow. */
         printf("  ");
-        print_uuid_part(&lflow->header_.uuid, print_uuid);
+        print_uuid_part(&curr->lflow->header_.uuid, print_uuid);
         printf("table=%-2"PRId64"(%-19s), priority=%-5"PRId64
                ", match=(%s), action=(%s)\n",
-               lflow->table_id,
-               smap_get_def(&lflow->external_ids, "stage-name", ""),
-               lflow->priority, lflow->match, lflow->actions);
+               curr->lflow->table_id,
+               smap_get_def(&curr->lflow->external_ids, "stage-name", ""),
+               curr->lflow->priority, curr->lflow->match,
+               curr->lflow->actions);
         if (vconn) {
-            sbctl_dump_openflow(vconn, &lflow->header_.uuid, stats);
+            sbctl_dump_openflow(vconn, &curr->lflow->header_.uuid, stats);
         }
-        prev = lflow;
+        prev = curr;
     }
 
     bool vflows = shash_find(&ctx->options, "--vflows") != NULL;
@@ -1119,6 +1230,7 @@ cmd_lflow_list(struct ctl_context *ctx)
         cmd_lflow_list_mac_bindings(ctx, vconn, datapath, stats, print_uuid);
         cmd_lflow_list_mc_groups(ctx, vconn, datapath, stats, print_uuid);
         cmd_lflow_list_chassis(ctx, vconn, stats, print_uuid);
+        cmd_lflow_list_load_balancers(ctx, vconn, datapath, stats, print_uuid);
     }
 
     vconn_close(vconn);
@@ -1410,11 +1522,41 @@ static const struct ctl_table_class tables[SBREC_N_TABLES] = {
     [SBREC_TABLE_ADDRESS_SET].row_ids[0]
     = {&sbrec_address_set_col_name, NULL, NULL},
 
+    [SBREC_TABLE_PORT_GROUP].row_ids[0]
+    = {&sbrec_port_group_col_name, NULL, NULL},
+
     [SBREC_TABLE_HA_CHASSIS_GROUP].row_ids[0]
     = {&sbrec_ha_chassis_group_col_name, NULL, NULL},
 
     [SBREC_TABLE_HA_CHASSIS].row_ids[0]
     = {&sbrec_ha_chassis_col_chassis, NULL, NULL},
+
+    [SBREC_TABLE_METER].row_ids[0]
+    = {&sbrec_meter_col_name, NULL, NULL},
+
+    [SBREC_TABLE_SERVICE_MONITOR].row_ids[0]
+    = {&sbrec_service_monitor_col_logical_port, NULL, NULL},
+
+    [SBREC_TABLE_DHCP_OPTIONS].row_ids[0]
+    = {&sbrec_dhcp_options_col_name, NULL, NULL},
+
+    [SBREC_TABLE_DHCPV6_OPTIONS].row_ids[0]
+    = {&sbrec_dhcpv6_options_col_name, NULL, NULL},
+
+    [SBREC_TABLE_CONNECTION].row_ids[0]
+    = {&sbrec_connection_col_target, NULL, NULL},
+
+    [SBREC_TABLE_RBAC_ROLE].row_ids[0]
+    = {&sbrec_rbac_role_col_name, NULL, NULL},
+
+    [SBREC_TABLE_RBAC_PERMISSION].row_ids[0]
+    = {&sbrec_rbac_permission_col_table, NULL, NULL},
+
+    [SBREC_TABLE_GATEWAY_CHASSIS].row_ids[0]
+    = {&sbrec_gateway_chassis_col_name, NULL, NULL},
+
+    [SBREC_TABLE_LOAD_BALANCER].row_ids[0]
+    = {&sbrec_load_balancer_col_name, NULL, NULL},
 };
 
 
