@@ -115,8 +115,8 @@ main(int argc, char *argv[])
                       "(use --help for help)");
         }
     } else {
-        if (argc != 2) {
-            ovs_fatal(0, "exactly two non-option arguments are required "
+        if (argc != 1 && argc != 2) {
+            ovs_fatal(0, "one or two non-option arguments are required "
                       "(use --help for help)");
         }
     }
@@ -134,8 +134,8 @@ main(int argc, char *argv[])
             ovs_fatal(error, "failed to create unixctl server");
         }
         unixctl_command_register("exit", "", 0, 0, ovntrace_exit, &exiting);
-        unixctl_command_register("trace", "[OPTIONS] DATAPATH MICROFLOW",
-                                 2, INT_MAX, ovntrace_trace, NULL);
+        unixctl_command_register("trace", "[OPTIONS] [DATAPATH] MICROFLOW",
+                                 1, INT_MAX, ovntrace_trace, NULL);
     }
     ovnsb_idl = ovsdb_idl_create(db, &sbrec_idl_class, true, false);
 
@@ -157,7 +157,9 @@ main(int argc, char *argv[])
 
             daemonize_complete();
             if (!get_detach()) {
-                char *output = trace(argv[0], argv[1]);
+                const char *dp_s = argc > 1 ? argv[0] : NULL;
+                const char *flow_s = argv[argc - 1];
+                char *output = trace(dp_s, flow_s);
                 fputs(output, stdout);
                 free(output);
                 return 0;
@@ -360,7 +362,7 @@ usage(void)
 {
     printf("\
 %s: OVN trace utility\n\
-usage: %s [OPTIONS] DATAPATH MICROFLOW\n\
+usage: %s [OPTIONS] [DATAPATH] MICROFLOW\n\
        %s [OPTIONS] --detach\n\
 \n\
 Output format options:\n\
@@ -872,18 +874,14 @@ ovntrace_make_names_friendly(const char *in)
 }
 
 static void
-read_flows(void)
+parse_lflow_for_datapath(const struct sbrec_logical_flow *sblf,
+                        const struct sbrec_datapath_binding *sbdb)
 {
-    ovn_init_symtab(&symtab);
-
-    const struct sbrec_logical_flow *sblf;
-    SBREC_LOGICAL_FLOW_FOR_EACH (sblf, ovnsb_idl) {
-        const struct sbrec_datapath_binding *sbdb = sblf->logical_datapath;
         struct ovntrace_datapath *dp
             = ovntrace_datapath_find_by_sb_uuid(&sbdb->header_.uuid);
         if (!dp) {
             VLOG_WARN("logical flow missing datapath");
-            continue;
+            return;
         }
 
         char *error;
@@ -895,7 +893,7 @@ read_flows(void)
             VLOG_WARN("%s: parsing expression failed (%s)",
                       sblf->match, error);
             free(error);
-            continue;
+            return;
         }
 
         struct ovnact_parse_params pp = {
@@ -917,7 +915,7 @@ read_flows(void)
             VLOG_WARN("%s: parsing actions failed (%s)", sblf->actions, error);
             free(error);
             expr_destroy(match);
-            continue;
+            return;
         }
 
         match = expr_combine(EXPR_T_AND, match, prereqs);
@@ -928,7 +926,7 @@ read_flows(void)
             expr_destroy(match);
             ovnacts_free(ovnacts.data, ovnacts.size);
             ofpbuf_uninit(&ovnacts);
-            continue;
+            return;
         }
         if (match) {
             match = expr_simplify(match);
@@ -958,6 +956,30 @@ read_flows(void)
                                    sizeof *dp->flows);
         }
         dp->flows[dp->n_flows++] = flow;
+}
+
+static void
+read_flows(void)
+{
+    ovn_init_symtab(&symtab);
+
+    const struct sbrec_logical_flow *sblf;
+    SBREC_LOGICAL_FLOW_FOR_EACH (sblf, ovnsb_idl) {
+        bool missing_datapath = true;
+
+        if (sblf->logical_datapath) {
+            parse_lflow_for_datapath(sblf, sblf->logical_datapath);
+            missing_datapath = false;
+        }
+
+        const struct sbrec_logical_dp_group *g = sblf->logical_dp_group;
+        for (size_t i = 0; g && i < g->n_datapaths; i++) {
+            parse_lflow_for_datapath(sblf, g->datapaths[i]);
+            missing_datapath = false;
+        }
+        if (missing_datapath) {
+            VLOG_WARN("logical flow missing datapath");
+        }
     }
 
     const struct ovntrace_datapath *dp;
@@ -1638,16 +1660,23 @@ execute_nd_ns(const struct ovnact_nest *on, const struct ovntrace_datapath *dp,
 static void
 execute_icmp4(const struct ovnact_nest *on,
               const struct ovntrace_datapath *dp,
-              const struct flow *uflow, uint8_t table_id,
+              const struct flow *uflow, uint8_t table_id, bool loopback,
               enum ovnact_pipeline pipeline, struct ovs_list *super)
 {
     struct flow icmp4_flow = *uflow;
 
     /* Update fields for ICMP. */
-    icmp4_flow.dl_dst = uflow->dl_dst;
-    icmp4_flow.dl_src = uflow->dl_src;
-    icmp4_flow.nw_dst = uflow->nw_dst;
-    icmp4_flow.nw_src = uflow->nw_src;
+    if (loopback) {
+        icmp4_flow.dl_dst = uflow->dl_src;
+        icmp4_flow.dl_src = uflow->dl_dst;
+        icmp4_flow.nw_dst = uflow->nw_src;
+        icmp4_flow.nw_src = uflow->nw_dst;
+    } else {
+        icmp4_flow.dl_dst = uflow->dl_dst;
+        icmp4_flow.dl_src = uflow->dl_src;
+        icmp4_flow.nw_dst = uflow->nw_dst;
+        icmp4_flow.nw_src = uflow->nw_src;
+    }
     icmp4_flow.nw_proto = IPPROTO_ICMP;
     icmp4_flow.nw_ttl = 255;
     icmp4_flow.tp_src = htons(ICMP4_DST_UNREACH); /* icmp type */
@@ -1663,16 +1692,23 @@ execute_icmp4(const struct ovnact_nest *on,
 static void
 execute_icmp6(const struct ovnact_nest *on,
               const struct ovntrace_datapath *dp,
-              const struct flow *uflow, uint8_t table_id,
+              const struct flow *uflow, uint8_t table_id, bool loopback,
               enum ovnact_pipeline pipeline, struct ovs_list *super)
 {
     struct flow icmp6_flow = *uflow;
 
     /* Update fields for ICMPv6. */
-    icmp6_flow.dl_dst = uflow->dl_dst;
-    icmp6_flow.dl_src = uflow->dl_src;
-    icmp6_flow.ipv6_dst = uflow->ipv6_dst;
-    icmp6_flow.ipv6_src = uflow->ipv6_src;
+    if (loopback) {
+        icmp6_flow.dl_dst = uflow->dl_src;
+        icmp6_flow.dl_src = uflow->dl_dst;
+        icmp6_flow.ipv6_dst = uflow->ipv6_src;
+        icmp6_flow.ipv6_src = uflow->ipv6_dst;
+    } else {
+        icmp6_flow.dl_dst = uflow->dl_dst;
+        icmp6_flow.dl_src = uflow->dl_src;
+        icmp6_flow.ipv6_dst = uflow->ipv6_dst;
+        icmp6_flow.ipv6_src = uflow->ipv6_src;
+    }
     icmp6_flow.nw_proto = IPPROTO_ICMPV6;
     icmp6_flow.nw_ttl = 255;
     icmp6_flow.tp_src = htons(ICMP6_DST_UNREACH); /* icmp type */
@@ -1686,18 +1722,26 @@ execute_icmp6(const struct ovnact_nest *on,
 }
 
 static void
-execute_tcp_reset(const struct ovnact_nest *on,
-                  const struct ovntrace_datapath *dp,
-                  const struct flow *uflow, uint8_t table_id,
-                  enum ovnact_pipeline pipeline, struct ovs_list *super)
+execute_tcp4_reset(const struct ovnact_nest *on,
+                   const struct ovntrace_datapath *dp,
+                   const struct flow *uflow, uint8_t table_id,
+                   bool loopback, enum ovnact_pipeline pipeline,
+                   struct ovs_list *super)
 {
     struct flow tcp_flow = *uflow;
 
     /* Update fields for TCP segment. */
-    tcp_flow.dl_dst = uflow->dl_dst;
-    tcp_flow.dl_src = uflow->dl_src;
-    tcp_flow.nw_dst = uflow->nw_dst;
-    tcp_flow.nw_src = uflow->nw_src;
+    if (loopback) {
+        tcp_flow.dl_dst = uflow->dl_src;
+        tcp_flow.dl_src = uflow->dl_dst;
+        tcp_flow.nw_dst = uflow->nw_src;
+        tcp_flow.nw_src = uflow->nw_dst;
+    } else {
+        tcp_flow.dl_dst = uflow->dl_dst;
+        tcp_flow.dl_src = uflow->dl_src;
+        tcp_flow.nw_dst = uflow->nw_dst;
+        tcp_flow.nw_src = uflow->nw_src;
+    }
     tcp_flow.nw_proto = IPPROTO_TCP;
     tcp_flow.nw_ttl = 255;
     tcp_flow.tp_src = uflow->tp_src;
@@ -1711,6 +1755,70 @@ execute_tcp_reset(const struct ovnact_nest *on,
                   table_id, pipeline, &node->subs);
 }
 
+static void
+execute_tcp6_reset(const struct ovnact_nest *on,
+                   const struct ovntrace_datapath *dp,
+                   const struct flow *uflow, uint8_t table_id,
+                   bool loopback, enum ovnact_pipeline pipeline,
+                   struct ovs_list *super)
+{
+    struct flow tcp_flow = *uflow;
+
+    /* Update fields for TCP segment. */
+    if (loopback) {
+        tcp_flow.dl_dst = uflow->dl_src;
+        tcp_flow.dl_src = uflow->dl_dst;
+        tcp_flow.ipv6_dst = uflow->ipv6_src;
+        tcp_flow.ipv6_src = uflow->ipv6_dst;
+    } else {
+        tcp_flow.dl_dst = uflow->dl_dst;
+        tcp_flow.dl_src = uflow->dl_src;
+        tcp_flow.ipv6_dst = uflow->ipv6_dst;
+        tcp_flow.ipv6_src = uflow->ipv6_src;
+    }
+    tcp_flow.nw_proto = IPPROTO_TCP;
+    tcp_flow.nw_ttl = 255;
+    tcp_flow.tp_src = uflow->tp_src;
+    tcp_flow.tp_dst = uflow->tp_dst;
+    tcp_flow.tcp_flags = htons(TCP_RST);
+
+    struct ovntrace_node *node = ovntrace_node_append(
+        super, OVNTRACE_NODE_TRANSFORMATION, "tcp_reset");
+
+    trace_actions(on->nested, on->nested_len, dp, &tcp_flow,
+                  table_id, pipeline, &node->subs);
+}
+
+static void
+execute_tcp_reset(const struct ovnact_nest *on,
+                  const struct ovntrace_datapath *dp,
+                  const struct flow *uflow, uint8_t table_id,
+                  bool loopback, enum ovnact_pipeline pipeline,
+                  struct ovs_list *super)
+{
+    if (get_dl_type(uflow) == htons(ETH_TYPE_IP)) {
+        execute_tcp4_reset(on, dp, uflow, table_id, loopback, pipeline, super);
+    } else {
+        execute_tcp6_reset(on, dp, uflow, table_id, loopback, pipeline, super);
+    }
+}
+static void
+execute_reject(const struct ovnact_nest *on,
+               const struct ovntrace_datapath *dp,
+               const struct flow *uflow, uint8_t table_id,
+               enum ovnact_pipeline pipeline, struct ovs_list *super)
+{
+    if (uflow->nw_proto == IPPROTO_TCP) {
+        execute_tcp_reset(on, dp, uflow, table_id, true, pipeline, super);
+    } else {
+        if (get_dl_type(uflow) == htons(ETH_TYPE_IP)) {
+            execute_icmp4(on, dp, uflow, table_id, true, pipeline, super);
+        } else {
+            execute_icmp6(on, dp, uflow, table_id, true, pipeline, super);
+        }
+    }
+}
+
 static void
 execute_get_mac_bind(const struct ovnact_get_mac_bind *bind,
                      const struct ovntrace_datapath *dp,
@@ -1888,12 +1996,15 @@ execute_next(const struct ovnact_next *next,
              enum ovnact_pipeline pipeline, struct ovs_list *super)
 {
     if (pipeline != next->pipeline) {
-        ovs_assert(next->pipeline == OVNACT_P_INGRESS);
-
-        uint16_t in_key = uflow->regs[MFF_LOG_INPORT - MFF_REG0];
+        uint16_t key = next->pipeline == OVNACT_P_INGRESS
+                       ? uflow->regs[MFF_LOG_INPORT - MFF_REG0]
+                       : uflow->regs[MFF_LOG_OUTPORT - MFF_REG0];
         struct ovntrace_node *node = ovntrace_node_append(
-            super, OVNTRACE_NODE_PIPELINE, "ingress(dp=\"%s\", inport=\"%s\")",
-            dp->friendly_name, ovntrace_port_key_to_name(dp, in_key));
+            super, OVNTRACE_NODE_PIPELINE, "%s(dp=\"%s\", %s=\"%s\")",
+            next->pipeline == OVNACT_P_INGRESS ? "ingress" : "egress",
+            dp->friendly_name,
+            next->pipeline == OVNACT_P_INGRESS ? "inport" : "outport",
+            ovntrace_port_key_to_name(dp, key));
         super = &node->subs;
     }
     trace__(dp, uflow, next->ltable, next->pipeline, super);
@@ -1901,7 +2012,7 @@ execute_next(const struct ovnact_next *next,
 
 
 static void
-execute_dns_lookup(const struct ovnact_dns_lookup *dl, struct flow *uflow,
+execute_dns_lookup(const struct ovnact_result *dl, struct flow *uflow,
                    struct ovs_list *super)
 {
     struct mf_subfield sf = expr_resolve_field(&dl->dst);
@@ -2133,6 +2244,57 @@ execute_ovnfield_load(const struct ovnact_load *load,
     }
 }
 
+static void
+execute_chk_lb_hairpin(const struct ovnact_result *dl, struct flow *uflow,
+                       struct ovs_list *super)
+{
+    int family = (uflow->dl_type == htons(ETH_TYPE_IP) ? AF_INET
+                  : uflow->dl_type == htons(ETH_TYPE_IPV6) ? AF_INET6
+                  : AF_UNSPEC);
+    uint8_t res = 0;
+    if (family != AF_UNSPEC && uflow->ct_state & CS_DST_NAT) {
+        if (family == AF_INET) {
+            res = (uflow->nw_src == uflow->nw_dst) ? 1 : 0;
+        } else {
+            res = ipv6_addr_equals(&uflow->ipv6_src, &uflow->ipv6_dst) ? 1 : 0;
+        }
+    }
+
+    struct mf_subfield sf = expr_resolve_field(&dl->dst);
+    union mf_subvalue sv = { .u8_val = res };
+    mf_write_subfield_flow(&sf, &sv, uflow);
+
+    struct ds s = DS_EMPTY_INITIALIZER;
+    expr_field_format(&dl->dst, &s);
+    ovntrace_node_append(super, OVNTRACE_NODE_MODIFY,
+                         "%s = %d", ds_cstr(&s), res);
+    ds_destroy(&s);
+}
+
+static void
+execute_chk_lb_hairpin_reply(const struct ovnact_result *dl,
+                             struct flow *uflow,
+                             struct ovs_list *super)
+{
+    struct mf_subfield sf = expr_resolve_field(&dl->dst);
+    union mf_subvalue sv = { .u8_val = 0 };
+    mf_write_subfield_flow(&sf, &sv, uflow);
+    ovntrace_node_append(super, OVNTRACE_NODE_ERROR,
+                         "*** chk_lb_hairpin_reply action not implemented");
+    struct ds s = DS_EMPTY_INITIALIZER;
+    expr_field_format(&dl->dst, &s);
+    ovntrace_node_append(super, OVNTRACE_NODE_MODIFY,
+                         "%s = 0", ds_cstr(&s));
+    ds_destroy(&s);
+}
+
+static void
+execute_ct_snat_to_vip(struct flow *uflow OVS_UNUSED, struct ovs_list *super)
+{
+    ovntrace_node_append(super, OVNTRACE_NODE_ERROR,
+                         "*** ct_snat_to_vip action not implemented");
+}
+
 static void
 trace_actions(const struct ovnact *ovnacts, size_t ovnacts_len,
               const struct ovntrace_datapath *dp, struct flow *uflow,
@@ -2194,7 +2356,8 @@ trace_actions(const struct ovnact *ovnacts, size_t ovnacts_len,
             execute_ct_next(ovnact_get_CT_NEXT(a), dp, uflow, pipeline, super);
             break;
 
-        case OVNACT_CT_COMMIT:
+        case OVNACT_CT_COMMIT_V1:
+        case OVNACT_CT_COMMIT_V2:
             /* Nothing to do. */
             break;
 
@@ -2312,23 +2475,23 @@ trace_actions(const struct ovnact *ovnacts, size_t ovnacts_len,
             break;
 
         case OVNACT_ICMP4:
-            execute_icmp4(ovnact_get_ICMP4(a), dp, uflow, table_id, pipeline,
-                          super);
+            execute_icmp4(ovnact_get_ICMP4(a), dp, uflow, table_id, false,
+                          pipeline, super);
             break;
 
         case OVNACT_ICMP4_ERROR:
             execute_icmp4(ovnact_get_ICMP4_ERROR(a), dp, uflow, table_id,
-                          pipeline, super);
+                          false, pipeline, super);
             break;
 
         case OVNACT_ICMP6:
-            execute_icmp6(ovnact_get_ICMP6(a), dp, uflow, table_id, pipeline,
-                          super);
+            execute_icmp6(ovnact_get_ICMP6(a), dp, uflow, table_id, false,
+                          pipeline, super);
             break;
 
         case OVNACT_ICMP6_ERROR:
             execute_icmp6(ovnact_get_ICMP6_ERROR(a), dp, uflow, table_id,
-                          pipeline, super);
+                          false, pipeline, super);
             break;
 
         case OVNACT_IGMP:
@@ -2337,13 +2500,30 @@ trace_actions(const struct ovnact *ovnacts, size_t ovnacts_len,
 
         case OVNACT_TCP_RESET:
             execute_tcp_reset(ovnact_get_TCP_RESET(a), dp, uflow, table_id,
-                              pipeline, super);
+                              false, pipeline, super);
             break;
 
         case OVNACT_OVNFIELD_LOAD:
             execute_ovnfield_load(ovnact_get_OVNFIELD_LOAD(a), super);
             break;
 
+        case OVNACT_REJECT:
+            execute_reject(ovnact_get_REJECT(a), dp, uflow, table_id,
+                           pipeline, super);
+            break;
+
+        case OVNACT_CHK_LB_HAIRPIN:
+            execute_chk_lb_hairpin(ovnact_get_CHK_LB_HAIRPIN(a), uflow, super);
+            break;
+
+        case OVNACT_CHK_LB_HAIRPIN_REPLY:
+            execute_chk_lb_hairpin_reply(ovnact_get_CHK_LB_HAIRPIN_REPLY(a),
+                                         uflow, super);
+            break;
+        case OVNACT_CT_SNAT_TO_VIP:
+            execute_ct_snat_to_vip(uflow, super);
+            break;
+
         case OVNACT_TRIGGER_EVENT:
             break;
 
@@ -2465,24 +2645,76 @@ trace__(const struct ovntrace_datapath *dp, struct flow *uflow,
     }
 }
 
+static char * OVS_WARN_UNUSED_RESULT
+trace_parse_error(char *error)
+{
+    char *s = xasprintf("error parsing flow: %s\n", error);
+    free(error);
+    return s;
+}
+
+static char * OVS_WARN_UNUSED_RESULT
+trace_parse(const char *dp_s, const char *flow_s,
+            const struct ovntrace_datapath **dpp, struct flow *uflow)
+{
+    *dpp = NULL;
+    if (dp_s) {
+        /* Use the specified datapath. */
+        *dpp = ovntrace_datapath_find_by_name(dp_s);
+        if (!dpp) {
+            return xasprintf("unknown datapath \"%s\"\n", dp_s);
+        }
+    } else {
+        /* Figure out the datapath from the flow, based on its inport.
+         *
+         * First make sure that the expression parses. */
+        char *error;
+        struct expr *e = expr_parse_string(flow_s, &symtab, &address_sets,
+                                           &port_groups, NULL, NULL, 0,
+                                           &error);
+        if (!e) {
+            return trace_parse_error(error);
+        }
+
+        /* Extract the name of the inport. */
+        char *port_name = expr_find_inport(e, &error);
+        expr_destroy(e);
+        if (!port_name) {
+            return trace_parse_error(error);
+        }
+
+        /* Find the port by name. */
+        const struct ovntrace_port *port = shash_find_data(&ports, port_name);
+        if (!port) {
+            char *s = xasprintf("unknown port \"%s\"\n", port_name);
+            free(port_name);
+            return s;
+        }
+
+        /* Use the port's datapath. */
+        *dpp = port->dp;
+        free(port_name);
+    }
+
+    char *error = expr_parse_microflow(flow_s, &symtab, &address_sets,
+                                       &port_groups, ovntrace_lookup_port,
+                                       *dpp, uflow);
+    if (error) {
+        return trace_parse_error(error);
+    }
+
+    return NULL;
+}
+
 static char *
 trace(const char *dp_s, const char *flow_s)
 {
-    const struct ovntrace_datapath *dp = ovntrace_datapath_find_by_name(dp_s);
-    if (!dp) {
-        return xasprintf("unknown datapath \"%s\"\n", dp_s);
-    }
-
+    const struct ovntrace_datapath *dp;
     struct flow uflow;
-    char *error = expr_parse_microflow(flow_s, &symtab, &address_sets,
-                                       &port_groups, ovntrace_lookup_port,
-                                       dp, &uflow);
+    char *error = trace_parse(dp_s, flow_s, &dp, &uflow);
     if (error) {
-        char *s = xasprintf("error parsing flow: %s\n", error);
-        free(error);
-        return s;
+        return error;
     }
-
     uint32_t in_key = uflow.regs[MFF_LOG_INPORT - MFF_REG0];
     if (!in_key) {
         VLOG_WARN("microflow does not specify ingress port");
@@ -2577,13 +2809,15 @@ ovntrace_trace(struct unixctl_conn *conn, int argc,
         detailed = true;
     }
 
-    if (argc != 3) {
+    if (argc != 2 && argc != 3) {
         unixctl_command_reply_error(
-            conn, "exactly 2 non-option arguments are required");
+            conn, "one or two non-option arguments are required");
         return;
     }
 
-    char *output = trace(argv[1], argv[2]);
+    const char *dp_s = argc > 2 ? argv[1] : NULL;
+    const char *flow_s = argv[argc - 1];
+    char *output = trace(dp_s, flow_s);
     unixctl_command_reply(conn, output);
     free(output);
 }
